"""Message transport using :pypi:`aiokafka`."""
import asyncio
import typing
import abc
import sys
import traceback
import warnings

from collections import deque
from time import monotonic
from typing import (
    Any,
    Awaitable,
    Callable,
    ClassVar,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Set,
    Tuple,
    Type,
    cast,
    no_type_check,
)

import aiokafka
import aiokafka.abc
import opentracing
from aiokafka.consumer.group_coordinator import OffsetCommitRequest
from aiokafka.errors import (
    CommitFailedError,
    ConsumerStoppedError,
    IllegalStateError,
    KafkaError,
    NotControllerError,
    TopicAlreadyExistsError as TopicExistsError,
    for_code,
)
from aiokafka.structs import (
    OffsetAndMetadata,
    TopicPartition as _TopicPartition,
)
from aiokafka.util import parse_kafka_version
from aiokafka import TopicPartition
from aiokafka.coordinator.assignors.roundrobin import RoundRobinPartitionAssignor
from aiokafka.partitioner import DefaultPartitioner, murmur2
from aiokafka.protocol.admin import CreateTopicsRequest
from aiokafka.protocol.metadata import MetadataRequest_v1
from mode import Service, get_logger
from mode.utils.futures import StampedeWrapper
from mode.utils.objects import cached_property
from mode.utils import text
from mode.utils.times import Seconds, humanize_seconds_ago, want_seconds
from mode.utils.typing import Deque
from opentracing.ext import tags
from yarl import URL

from faust.auth import (
    GSSAPICredentials,
    SASLCredentials,
    SSLCredentials,
)
from faust.exceptions import (
    ConsumerNotStarted,
    ImproperlyConfigured,
    NotReady,
    ProducerSendError,
)
from faust.transport import base
from faust.transport.consumer import (
    ConsumerThread,
    RecordMap,
    ThreadDelegateConsumer,
    ensure_TPset,
)
from faust.types import ConsumerMessage, HeadersArg, RecordMetadata, TP
from faust.types.auth import CredentialsT
from faust.types.transports import (
    ConsumerT,
    PartitionerT,
    ProducerT,
)
from faust.utils.tracing import (
    noop_span,
    set_current_span,
    traced_from_parent_span,
)

from aiokafka.client import AIOKafkaClient
from aiokafka.codec import has_gzip, has_lz4, has_snappy, has_zstd
from aiokafka.errors import (
    IllegalOperation,
    MessageSizeTooLargeError,
    UnsupportedVersionError,
)

from aiokafka.record.default_records import DefaultRecordBatch
from aiokafka.record.legacy_records import LegacyRecordBatchBuilder
from aiokafka.util import (
    commit_structure_validate,
    create_task,
    get_running_loop,
)

from aiokafka.producer.message_accumulator import MessageAccumulator
from aiokafka.producer.sender import Sender
from aiokafka.producer.transaction_manager import TransactionManager
from aiokafka.producer import AIOKafkaProducer

_DEFAULT_PARTITIONER = DefaultPartitioner

_missing = object()

__all__ = ['Consumer', 'Producer', 'Transport']

logger = get_logger(__name__)

DEFAULT_GENERATION_ID = OffsetCommitRequest.DEFAULT_GENERATION_ID

TOPIC_LENGTH_MAX = 249

SLOW_PROCESSING_CAUSE_AGENT = '''
The agent processing the stream is hanging (waiting for network, I/O or \
infinite loop).
'''.strip()

SLOW_PROCESSING_CAUSE_STREAM = '''
The stream has stopped processing events for some reason.
'''.strip()


SLOW_PROCESSING_CAUSE_COMMIT = '''
The commit handler background thread has stopped working (report as bug).
'''.strip()


SLOW_PROCESSING_EXPLAINED = '''

There are multiple possible explanations for this:

1) The processing of a single event in the stream
   is taking too long.

    The timeout for this is defined by the %(setting)s setting,
    currently set to %(current_value)r.  If you expect the time
    required to process an event, to be greater than this then please
    increase the timeout.

'''

SLOW_PROCESSING_NO_FETCH_SINCE_START = '''
Aiokafka has not sent fetch request for %r since start (started %s)
'''.strip()

SLOW_PROCESSING_NO_RESPONSE_SINCE_START = '''
Aiokafka has not received fetch response for %r since start (started %s)
'''.strip()

SLOW_PROCESSING_NO_RECENT_FETCH = '''
Aiokafka stopped fetching from %r (last done %s)
'''.strip()

SLOW_PROCESSING_NO_RECENT_RESPONSE = '''
Broker stopped responding to fetch requests for %r (last responded %s)
'''.strip()

SLOW_PROCESSING_NO_HIGHWATER_SINCE_START = '''
Highwater not yet available for %r (started %s).
'''.strip()

SLOW_PROCESSING_STREAM_IDLE_SINCE_START = '''
Stream has not started processing %r (started %s).
'''.strip()

SLOW_PROCESSING_STREAM_IDLE = '''
Stream stopped processing, or is slow for %r (last inbound %s).
'''.strip()

SLOW_PROCESSING_NO_COMMIT_SINCE_START = '''
Has not committed %r at all since worker start (started %s).
'''.strip()

SLOW_PROCESSING_NO_RECENT_COMMIT = '''
Has not committed %r (last commit %s).
'''.strip()


def server_list(urls: List[URL], default_port: int) -> List[str]:
    """Convert list of urls to list of servers accepted by :pypi:`aiokafka`."""
    default_host = '127.0.0.1'
    return [f'{u.host or default_host}:{u.port or default_port}' for u in urls]


class ConsumerRebalanceListener(
        aiokafka.abc.ConsumerRebalanceListener):  # type: ignore
    # kafka's ridiculous class based callback interface makes this hacky.

    def __init__(self, thread: ConsumerThread) -> None:
        self._thread: ConsumerThread = thread

    def on_partitions_revoked(
            self, revoked: Iterable[_TopicPartition]) -> Awaitable:
        """Call when partitions are being revoked."""
        thread = self._thread
        # XXX Must call app.on_rebalance_start as early as possible.
        # we call this in the sync method, this way when we know
        # that it will be called even if await never returns to the coroutine.
        thread.app.on_rebalance_start()

        # this way we should also get a warning if the coroutine
        # is never awaited.
        return thread.on_partitions_revoked(ensure_TPset(revoked))

    async def on_partitions_assigned(
            self, assigned: Iterable[_TopicPartition]) -> None:
        """Call when partitions are being assigned."""
        await self._thread.on_partitions_assigned(ensure_TPset(assigned))


class Consumer(ThreadDelegateConsumer):
    """Kafka consumer using :pypi:`aiokafka`."""

    logger = logger

    RebalanceListener: ClassVar[Type[ConsumerRebalanceListener]]
    RebalanceListener = ConsumerRebalanceListener

    consumer_stopped_errors: ClassVar[Tuple[Type[BaseException], ...]] = (
        ConsumerStoppedError,
    )

    def _new_consumer_thread(self) -> ConsumerThread:
        return AIOKafkaConsumerThread(self, loop=self.loop, beacon=self.beacon)

    async def create_topic(self,
                           topic: str,
                           partitions: int,
                           replication: int,
                           *,
                           config: Mapping[str, Any] = None,
                           timeout: Seconds = 30.0,
                           retention: Seconds = None,
                           compacting: bool = None,
                           deleting: bool = None,
                           ensure_created: bool = False) -> None:
        """Create/declare topic on server."""
        await self._thread.create_topic(
            topic,
            partitions,
            replication,
            config=config,
            timeout=timeout,
            retention=retention,
            compacting=compacting,
            deleting=deleting,
            ensure_created=ensure_created,
        )

    def _new_topicpartition(self, topic: str, partition: int) -> TP:
        return cast(TP, _TopicPartition(topic, partition))

    def _to_message(self, tp: TP, record: Any) -> ConsumerMessage:
        timestamp: Optional[int] = record.timestamp
        timestamp_s: float = cast(float, None)
        if timestamp is not None:
            timestamp_s = timestamp / 1000.0
        return ConsumerMessage(
            record.topic,
            record.partition,
            record.offset,
            timestamp_s,
            record.timestamp_type,
            record.headers,
            record.key,
            record.value,
            record.checksum,
            record.serialized_key_size,
            record.serialized_value_size,
            tp,
        )

    async def on_stop(self) -> None:
        """Call when consumer is stopping."""
        await super().on_stop()
        transport = cast(Transport, self.transport)
        transport._topic_waiters.clear()


class AIOKafkaConsumerThread(ConsumerThread):
    _consumer: Optional[aiokafka.AIOKafkaConsumer] = None
    _pending_rebalancing_spans: Deque[opentracing.Span]

    tp_last_committed_at: MutableMapping[TP, float]
    time_started: float

    tp_fetch_request_timeout_secs: float
    tp_fetch_response_timeout_secs: float
    tp_stream_timeout_secs: float
    tp_commit_timeout_secs: float

    def __post_init__(self) -> None:
        consumer = cast(Consumer, self.consumer)
        self._partitioner: PartitionerT = (
            self.app.conf.producer_partitioner or DefaultPartitioner())
        self._rebalance_listener = consumer.RebalanceListener(self)
        self._pending_rebalancing_spans = deque()
        self.tp_last_committed_at = {}

        app = self.consumer.app

        stream_processing_timeout = app.conf.stream_processing_timeout
        self.tp_fetch_request_timeout_secs = stream_processing_timeout
        self.tp_fetch_response_timeout_secs = stream_processing_timeout
        self.tp_stream_timeout_secs = stream_processing_timeout

        commit_livelock_timeout = app.conf.broker_commit_livelock_soft_timeout
        self.tp_commit_timeout_secs = commit_livelock_timeout

    async def on_start(self) -> None:
        """Call when consumer starts."""
        self._consumer = self._create_consumer(loop=self.thread_loop)
        self.time_started = monotonic()
        await self._consumer.start()

    async def on_thread_stop(self) -> None:
        """Call when consumer thread is stopping."""
        # super stops thread method queue (QueueServiceThread.method_queue)
        await super().on_thread_stop()
        # when method queue is stopped, we can stop the consumer
        if self._consumer is not None:
            await self._consumer.stop()

    def _create_consumer(
            self,
            loop: asyncio.AbstractEventLoop) -> aiokafka.AIOKafkaConsumer:
        transport = cast(Transport, self.transport)
        if self.app.client_only:
            return self._create_client_consumer(transport, loop=loop)
        else:
            return self._create_worker_consumer(transport, loop=loop)

    def _create_worker_consumer(
            self,
            transport: 'Transport',
            loop: asyncio.AbstractEventLoop) -> aiokafka.AIOKafkaConsumer:
        isolation_level: str = 'read_uncommitted'
        conf = self.app.conf
        if self.consumer.in_transaction:
            isolation_level = 'read_committed'
        self._assignor = (
            self.app.assignor
            if self.app.conf.table_standby_replicas > 0
            else RoundRobinPartitionAssignor
        )
        auth_settings = credentials_to_aiokafka_auth(
            conf.broker_credentials, conf.ssl_context)
        max_poll_interval = conf.broker_max_poll_interval or 0

        request_timeout = conf.broker_request_timeout
        session_timeout = conf.broker_session_timeout
        rebalance_timeout = conf.broker_rebalance_timeout

        if session_timeout > request_timeout:
            raise ImproperlyConfigured(
                f'Setting broker_session_timeout={session_timeout} '
                f'cannot be greater than '
                f'broker_request_timeout={request_timeout}')

        return aiokafka.AIOKafkaConsumer(
            loop=loop,
            api_version=conf.consumer_api_version,
            client_id=conf.broker_client_id,
            group_id=conf.id,
            group_instance_id=conf.consumer_group_instance_id,
            bootstrap_servers=server_list(
                transport.url, transport.default_port),
            partition_assignment_strategy=[self._assignor],
            enable_auto_commit=False,
            auto_offset_reset=conf.consumer_auto_offset_reset,
            max_poll_records=conf.broker_max_poll_records,
            max_poll_interval_ms=int(max_poll_interval * 1000.0),
            max_partition_fetch_bytes=conf.consumer_max_fetch_size,
            fetch_max_wait_ms=1500,
            request_timeout_ms=int(request_timeout * 1000.0),
            check_crcs=conf.broker_check_crcs,
            session_timeout_ms=int(session_timeout * 1000.0),
            rebalance_timeout_ms=int(rebalance_timeout * 1000.0),
            heartbeat_interval_ms=int(conf.broker_heartbeat_interval * 1000.0),
            isolation_level=isolation_level,
            **auth_settings,
        )

    def _create_client_consumer(
            self,
            transport: 'Transport',
            loop: asyncio.AbstractEventLoop) -> aiokafka.AIOKafkaConsumer:
        conf = self.app.conf
        auth_settings = credentials_to_aiokafka_auth(
            conf.broker_credentials, conf.ssl_context)
        max_poll_interval = conf.broker_max_poll_interval or 0
        return aiokafka.AIOKafkaConsumer(
            client_id=conf.broker_client_id,
            bootstrap_servers=server_list(transport.url, transport.default_port),
            request_timeout_ms=int(conf.broker_request_timeout * 1000.0),
            enable_auto_commit=True,
            max_poll_records=conf.broker_max_poll_records,
            max_poll_interval_ms=int(max_poll_interval * 1000.0),
            auto_offset_reset=conf.consumer_auto_offset_reset,
            check_crcs=conf.broker_check_crcs,
            **auth_settings,
        )

    @cached_property
    def trace_category(self) -> str:
        return f'{self.app.conf.name}-_aiokafka'

    def start_rebalancing_span(self) -> opentracing.Span:
        return self._start_span('rebalancing', lazy=True)

    def start_coordinator_span(self) -> opentracing.Span:
        return self._start_span('coordinator')

    def _start_span(self, name: str, *,
                    lazy: bool = False) -> opentracing.Span:
        tracer = self.app.tracer
        if tracer is not None:
            span = tracer.get_tracer(self.trace_category).start_span(
                operation_name=name,
            )
            span.set_tag(tags.SAMPLING_PRIORITY, 1)
            self.app._span_add_default_tags(span)
            set_current_span(span)
            if lazy:
                self._transform_span_lazy(span)
            return span
        else:
            return noop_span()

    @no_type_check
    def _transform_span_lazy(self, span: opentracing.Span) -> None:
        # XXX slow
        consumer = self
        if typing.TYPE_CHECKING:
            # MyPy completely disallows the statements below
            # claiming it is an illegal dynamic baseclass.
            # We know mypy, but do it anyway :D
            pass
        else:
            cls = span.__class__
            class LazySpan(cls):

                def finish() -> None:
                    consumer._span_finish(span)

            span._real_finish, span.finish = span.finish, LazySpan.finish

    def _span_finish(self, span: opentracing.Span) -> None:
        assert self._consumer is not None
        if self._consumer._coordinator.generation == DEFAULT_GENERATION_ID:
            self._on_span_generation_pending(span)
        else:
            self._on_span_generation_known(span)

    def _on_span_generation_pending(self, span: opentracing.Span) -> None:
        self._pending_rebalancing_spans.append(span)

    def _on_span_generation_known(self, span: opentracing.Span) -> None:
        if self._consumer:
            coordinator = self._consumer._coordinator
            coordinator_id = coordinator.coordinator_id
            app_id = self.app.conf.id
            generation = coordinator.generation
            member_id = coordinator.member_id

            try:
                op_name = span.operation_name
                set_tag = span.set_tag
            except AttributeError:  # pragma: no cover
                pass  # not a real span
            else:
                trace_id_str = f'reb-{app_id}-{generation}'
                trace_id = murmur2(trace_id_str.encode())

                span.context.trace_id = trace_id
                if op_name.endswith('.REPLACE_WITH_MEMBER_ID'):
                    span.set_operation_name(f'rebalancing node {member_id}')
                set_tag('kafka_generation', generation)
                set_tag('kafka_member_id', member_id)
                set_tag('kafka_coordinator_id', coordinator_id)
                self.app._span_add_default_tags(span)
                span._real_finish()

    def _on_span_cancelled_early(self, span: opentracing.Span) -> None:
        try:
            op_name = span.operation_name
        except AttributeError:
            return
        else:
            span.set_operation_name(f'{op_name} (CANCELLED)')
            span._real_finish()

    def traced_from_parent_span(self,
                                parent_span: opentracing.Span,
                                lazy: bool = False,
                                **extra_context: Any) -> Callable:
        return traced_from_parent_span(
            parent_span,
            callback=self._transform_span_lazy if lazy else None,
            **extra_context)

    def flush_spans(self) -> None:
        while self._pending_rebalancing_spans:
            span = self._pending_rebalancing_spans.popleft()
            self._on_span_cancelled_early(span)

    def on_generation_id_known(self) -> None:
        while self._pending_rebalancing_spans:
            span = self._pending_rebalancing_spans.popleft()
            self._on_span_generation_known(span)

    def close(self) -> None:
        """Close consumer for graceful shutdown."""
        if self._consumer is not None:
            self._consumer._closed = True
            asyncio.run_coroutine_threadsafe(
                self._consumer._client.close(), self.app.loop
            )
            asyncio.run_coroutine_threadsafe(
                self._consumer._coordinator.close(), self.app.loop
            )            

    async def subscribe(self, topics: Iterable[str]) -> None:
        """Reset subscription (requires rebalance)."""
        # XXX pattern does not work :/
        await self.call_thread(
            self._ensure_consumer().subscribe,
            topics=set(topics),
            listener=self._rebalance_listener,
        )

    async def seek_to_committed(self) -> Mapping[TP, int]:
        """Seek partitions to the last committed offset."""
        return await self.call_thread(
            self._ensure_consumer().seek_to_committed)

    async def commit(self, offsets: Mapping[TP, int]) -> bool:
        """Commit topic offsets."""
        return await self.call_thread(self._commit, offsets)

    async def _commit(self, offsets: Mapping[TP, int]) -> bool:
        consumer = self._ensure_consumer()
        now = monotonic()
        try:
            aiokafka_offsets = {
                tp: OffsetAndMetadata(offset, '')
                for tp, offset in offsets.items()
            }
            self.tp_last_committed_at.update({
                tp: now
                for tp in offsets
            })
            await consumer.commit(aiokafka_offsets)
        except CommitFailedError as exc:
            if 'already rebalanced' in str(exc):
                return False
            self.log.exception('Committing raised exception: %r', exc)
            await self.crash(exc)
            return False
        except IllegalStateError as exc:
            self.log.exception(
                'Got exception: %r\nCurrent assignment: %r',
                exc, self.assignment())
            await self.crash(exc)
            return False
        return True

    def verify_event_path(self, now: float, tp: TP) -> None:
        # long function ahead, but not difficult to test
        # as it always returns as soon as some condition is met.
        if self._verify_aiokafka_event_path(now, tp):
            # already logged error.
            return None
        parent = cast(Consumer, self.consumer)
        app = parent.app
        monitor = app.monitor
        acks_enabled_for = app.topics.acks_enabled_for
        secs_since_started = now - self.time_started

        if monitor is not None:  # need for .stream_inbound_time
            highwater = self.highwater(tp)
            committed_offset = parent._committed_offset.get(tp)
            has_acks = acks_enabled_for(tp.topic)
            if highwater is None:
                if secs_since_started >= self.tp_stream_timeout_secs:
                    # AIOKAFKA HAS NOT UPDATED HIGHWATER SINCE STARTING
                    self.log.error(
                        SLOW_PROCESSING_NO_HIGHWATER_SINCE_START,
                        tp, humanize_seconds_ago(secs_since_started),
                    )
                return None

            if has_acks and committed_offset is not None:
                if highwater > committed_offset:
                    inbound = monitor.stream_inbound_time.get(tp)
                    if inbound is None:
                        if secs_since_started >= self.tp_stream_timeout_secs:
                            # AIOKAFKA IS FETCHING BUT STREAM IS NOT
                            # PROCESSING EVENTS (no events at all since
                            # start).
                            self._log_slow_processing_stream(
                                SLOW_PROCESSING_STREAM_IDLE_SINCE_START,
                                tp, humanize_seconds_ago(secs_since_started),
                            )
                        return None

                    secs_since_stream = now - inbound
                    if secs_since_stream >= self.tp_stream_timeout_secs:
                        # AIOKAFKA IS FETCHING, AND STREAM WAS WORKING
                        # BEFORE BUT NOW HAS STOPPED PROCESSING
                        # (or processing of an event in the stream takes
                        #  longer than tp_stream_timeout_secs).
                        self._log_slow_processing_stream(
                            SLOW_PROCESSING_STREAM_IDLE,
                            tp, humanize_seconds_ago(secs_since_stream),
                        )
                        return None

                    last_commit = self.tp_last_committed_at.get(tp)
                    if last_commit is None:
                        if secs_since_started >= self.tp_commit_timeout_secs:
                            # AIOKAFKA IS FETCHING AND STREAM IS PROCESSING
                            # BUT WE HAVE NOT COMMITTED ANYTHING SINCE WORKER
                            # START.
                            self._log_slow_processing_commit(
                                SLOW_PROCESSING_NO_COMMIT_SINCE_START,
                                tp, humanize_seconds_ago(secs_since_started),
                            )
                            return None
                    else:
                        secs_since_commit = now - last_commit
                        if secs_since_commit >= self.tp_commit_timeout_secs:
                            # AIOKAFKA IS FETCHING AND STREAM IS PROCESSING
                            # BUT WE HAVE NOT COMITTED ANYTHING IN A WHILE
                            # (commit offset is not advancing).
                            self._log_slow_processing_commit(
                                SLOW_PROCESSING_NO_RECENT_COMMIT,
                                tp, humanize_seconds_ago(secs_since_commit),
                            )
                            return None

    def verify_recovery_event_path(self, now: float, tp: TP) -> None:
        self._verify_aiokafka_event_path(now, tp)

    def _verify_aiokafka_event_path(self, now: float, tp: TP) -> bool:
        """Verify that :pypi:`aiokafka` event path is working.

        Returns :const:`True` if any error was logged.
        """
        consumer = self._ensure_consumer()
        secs_since_started = now - self.time_started
        aiotp = TopicPartition(tp.topic, tp.partition)
        assignment = consumer._fetcher._subscriptions.subscription.assignment
        if not assignment or not assignment.active:
            self.log.error(f"No active partitions for {tp}")
            return True
        poll_at = None
        aiotp_state = assignment.state_value(aiotp)
        if aiotp_state and aiotp_state.timestamp:
            poll_at = aiotp_state.timestamp / 1000  # milliseconds
        if poll_at is None:
            if secs_since_started >= self.tp_fetch_request_timeout_secs:
                # NO FETCH REQUEST SENT AT ALL SINCE WORKER START
                self.log.error(
                    SLOW_PROCESSING_NO_FETCH_SINCE_START,
                    tp,
                    humanize_seconds_ago(secs_since_started),
                )
            return True

        secs_since_request = now - poll_at
        if secs_since_request >= self.tp_fetch_request_timeout_secs:
            # NO REQUEST SENT BY AIOKAFKA IN THE LAST n SECONDS
            self.log.error(
                SLOW_PROCESSING_NO_RECENT_FETCH,
                tp,
                humanize_seconds_ago(secs_since_request),
            )
            return True

        return False

    def _log_slow_processing_stream(self, msg: str, *args: Any) -> None:
        app = self.consumer.app
        self._log_slow_processing(
            msg, *args,
            causes=[
                SLOW_PROCESSING_CAUSE_STREAM,
                SLOW_PROCESSING_CAUSE_AGENT,
            ],
            setting='stream_processing_timeout',
            current_value=app.conf.stream_processing_timeout,
        )

    def _log_slow_processing_commit(self, msg: str, *args: Any) -> None:
        app = self.consumer.app
        self._log_slow_processing(
            msg, *args,
            causes=[SLOW_PROCESSING_CAUSE_COMMIT],
            setting='broker_commit_livelock_soft_timeout',
            current_value=app.conf.broker_commit_livelock_soft_timeout,
        )

    def _make_slow_processing_error(self,
                                    msg: str,
                                    causes: Iterable[str]) -> str:
        return ' '.join([
            msg,
            SLOW_PROCESSING_EXPLAINED,
            text.enumeration(causes, start=2, sep='\n\n'),
        ])

    def _log_slow_processing(self, msg: str, *args: Any,
                             causes: Iterable[str],
                             setting: str,
                             current_value: float) -> None:
        return self.log.error(
            self._make_slow_processing_error(msg, causes),
            *args,
            setting=setting,
            current_value=current_value,
        )

    async def position(self, tp: TP) -> Optional[int]:
        """Return the current position for topic partition."""
        return await self.call_thread(
            self._ensure_consumer().position, tp)

    async def seek_to_beginning(self, *partitions: _TopicPartition) -> None:
        """Seek list of offsets to the first available offset."""
        await self.call_thread(
            self._ensure_consumer().seek_to_beginning, *partitions)

    async def seek_wait(self, partitions: Mapping[TP, int]) -> None:
        """Seek partitions to specific offset and wait for operation."""
        consumer = self._ensure_consumer()
        await self.call_thread(self._seek_wait, consumer, partitions)

    async def _seek_wait(self,
                         consumer: Consumer,
                         partitions: Mapping[TP, int]) -> None:
        for tp, offset in partitions.items():
            self.log.dev('SEEK %r -> %r', tp, offset)
            consumer.seek(tp, offset)
            if offset > 0:
                self.consumer._read_offset[tp] = offset
        await asyncio.gather(*[
            consumer.position(tp) for tp in partitions
        ])

    def seek(self, partition: TP, offset: int) -> None:
        """Seek partition to specific offset."""
        self._ensure_consumer().seek(partition, offset)

    def assignment(self) -> Set[TP]:
        """Return the current assignment."""
        return ensure_TPset(self._ensure_consumer().assignment())

    def highwater(self, tp: TP) -> int:
        """Return the last offset in a specific partition."""
        if self.consumer.in_transaction:
            return self._ensure_consumer().last_stable_offset(tp)
        else:
            return self._ensure_consumer().highwater(tp)

    def topic_partitions(self, topic: str) -> Optional[int]:
        """Return the number of partitions configured for topic by name."""
        if self._consumer is not None:
            return self._consumer._coordinator._metadata_snapshot.get(topic)
        return None

    async def earliest_offsets(self,
                               *partitions: TP) -> Mapping[TP, int]:
        """Return the earliest offsets for a list of partitions."""
        return await self.call_thread(
            self._ensure_consumer().beginning_offsets, partitions)

    async def highwaters(self, *partitions: TP) -> Mapping[TP, int]:
        """Return the last offsets for a list of partitions."""
        return await self.call_thread(self._highwaters, partitions)

    async def _highwaters(self, partitions: List[TP]) -> Mapping[TP, int]:
        consumer = self._ensure_consumer()
        if self.consumer.in_transaction:
            return {
                tp: consumer.last_stable_offset(tp)
                for tp in partitions
            }
        else:
            return cast(Mapping[TP, int],
                        await consumer.end_offsets(partitions))

    def _ensure_consumer(self) -> aiokafka.AIOKafkaConsumer:
        if self._consumer is None:
            raise ConsumerNotStarted('Consumer thread not yet started')
        return self._consumer

    async def getmany(self,
                      active_partitions: Optional[Set[TP]],
                      timeout: float) -> RecordMap:
        """Fetch batch of messages from server."""
        # Implementation for the Fetcher service.
        _consumer = self._ensure_consumer()
        # NOTE: Since we are enqueing the fetch request,
        # we need to check when dequeued that we are not in a rebalancing
        # state at that point to return early, or we
        # will create a deadlock (fetch request starts after flow stopped)
        return await self.call_thread(
            self._fetch_records,
            _consumer,
            active_partitions,
            timeout=timeout,
            max_records=_consumer._max_poll_records,
        )

    async def _fetch_records(self,
                             consumer: aiokafka.AIOKafkaConsumer,
                             active_partitions: Set[TP],
                             timeout: float = None,
                             max_records: int = None) -> RecordMap:
        if not self.consumer.flow_active:
            return {}
        fetcher = consumer._fetcher
        if consumer._closed or fetcher._closed:
            raise ConsumerStoppedError()
        with fetcher._subscriptions.fetch_context():
            return await fetcher.fetched_records(
                active_partitions,
                timeout=timeout,
                max_records=max_records,
            )      

    async def create_topic(self,
                           topic: str,
                           partitions: int,
                           replication: int,
                           *,
                           config: Mapping[str, Any] = None,
                           timeout: Seconds = 30.0,
                           retention: Seconds = None,
                           compacting: bool = None,
                           deleting: bool = None,
                           ensure_created: bool = False) -> None:
        """Create/declare topic on server."""
        transport = cast(Transport, self.consumer.transport)
        _consumer = self._ensure_consumer()
        _retention = (int(want_seconds(retention) * 1000.0)
                      if retention else None)
        if len(topic) > TOPIC_LENGTH_MAX:
            raise ValueError(
                f'Topic name {topic!r} is too long (max={TOPIC_LENGTH_MAX})')
        await self.call_thread(
            transport._create_topic,
            self,
            _consumer._client,
            topic,
            partitions,
            replication,
            config=config,
            timeout=int(want_seconds(timeout) * 1000.0),
            retention=_retention,
            compacting=compacting,
            deleting=deleting,
            ensure_created=ensure_created,
        )

    def key_partition(self,
                      topic: str,
                      key: Optional[bytes],
                      partition: int = None) -> Optional[int]:
        """Hash key to determine partition destination."""
        consumer = self._ensure_consumer()
        metadata = consumer._client.cluster
        partitions_for_topic = metadata.partitions_for_topic(topic)
        if partitions_for_topic is None:
            return None
        if partition is not None:
            assert partition >= 0
            assert partition in partitions_for_topic, \
                'Unrecognized partition'
            return partition

        all_partitions = list(partitions_for_topic)
        available = list(metadata.available_partitions_for_topic(topic))
        return self._partitioner(key, all_partitions, available)
    
    def request_rejoin(self):
        """Force consumer to rejoin group."""
        self._consumer._coordinator.request_rejoin()


class BaseProducer(abc.ABC):

    log = logger

    _PRODUCER_CLIENT_ID_SEQUENCE = 0

    _COMPRESSORS = {
        "gzip": (has_gzip, DefaultRecordBatch.CODEC_GZIP),
        "snappy": (has_snappy, DefaultRecordBatch.CODEC_SNAPPY),
        "lz4": (has_lz4, DefaultRecordBatch.CODEC_LZ4),
        "zstd": (has_zstd, DefaultRecordBatch.CODEC_ZSTD),
    }

    _closed = None  # Serves as an uninitialized flag for __del__
    _source_traceback = None

    def __init__(self, 
                 *, 
                 loop, 
                 bootstrap_servers='localhost',
                 client_id=None,
                 metadata_max_age_ms=300000, 
                 request_timeout_ms=40000,
                 api_version='auto', 
                 acks=_missing,
                 key_serializer=None, 
                 value_serializer=None,
                 compression_type=None, 
                 max_batch_size=16384,
                 partitioner=_DEFAULT_PARTITIONER, 
                 max_request_size=1048576,
                 linger_ms=0, 
                 retry_backoff_ms=100, 
                 security_protocol="PLAINTEXT",
                 ssl_context=None, 
                 connections_max_idle_ms=540000,
                 enable_idempotence=False, 
                 transactional_id=None,
                 transaction_timeout_ms=60000, 
                 sasl_mechanism="PLAIN",
                 sasl_plain_password=None, 
                 sasl_plain_username=None,
                 sasl_kerberos_service_name='kafka',
                 sasl_kerberos_domain_name=None,
                 sasl_oauth_token_provider=None
    ):
        
        if loop is None:
            loop = get_running_loop()
        else:
            warnings.warn(
                "The loop argument is deprecated since 0.7.1, "
                "and scheduled for removal in 0.9.0",
                DeprecationWarning,
                stacklevel=2,
            )
        if loop.get_debug():
            self._source_traceback = traceback.extract_stack(sys._getframe(1))
        self._loop = loop
                
        if acks not in (0, 1, -1, 'all', _missing):
            raise ValueError("Invalid ACKS parameter")
        if compression_type not in ('gzip', 'snappy', 'lz4', None):
            raise ValueError("Invalid compression type!")
        if compression_type:
            checker, compression_attrs = self._COMPRESSORS[compression_type]
            if not checker():
                raise RuntimeError("Compression library for {} not found"
                                   .format(compression_type))
        else:
            compression_attrs = 0
        self._compression_attrs = compression_attrs

        if acks is _missing:
            acks = 1
        elif acks == 'all':
            acks = -1

        AIOKafkaProducer._PRODUCER_CLIENT_ID_SEQUENCE += 1
        if client_id is None:
            client_id = 'aiokafka-producer-%s' % \
                AIOKafkaProducer._PRODUCER_CLIENT_ID_SEQUENCE
            
        self._bootstrap_servers = bootstrap_servers
        self._client_id = client_id
        self._metadata_max_age_ms = metadata_max_age_ms
        self._request_timeout_ms = request_timeout_ms
        self._api_version = api_version
        self._acks = acks
        self._key_serializer = key_serializer
        self._value_serializer = value_serializer
        self._compression_type = compression_type
        self._max_batch_size = max_batch_size
        self._partitioner = partitioner
        self._max_request_size = max_request_size
        self._linger_ms = linger_ms
        self._retry_backoff_ms = retry_backoff_ms
        self._security_protocol = security_protocol
        self._ssl_context = ssl_context
        self._connections_max_idle_ms = connections_max_idle_ms
        self._transaction_timeout_ms = transaction_timeout_ms
        self._transaction_timeout_ms = transaction_timeout_ms
        self._sasl_mechanism = sasl_mechanism
        self._sasl_plain_username = sasl_plain_username
        self._sasl_plain_password = sasl_plain_password
        self._sasl_kerberos_service_name = sasl_kerberos_service_name
        self._sasl_kerberos_domain_name = sasl_kerberos_domain_name

        self.client = AIOKafkaClient(
            loop=loop, 
            bootstrap_servers=bootstrap_servers,
            client_id=client_id, 
            metadata_max_age_ms=metadata_max_age_ms,
            request_timeout_ms=request_timeout_ms,
            retry_backoff_ms=retry_backoff_ms,
            api_version=api_version, 
            security_protocol=security_protocol,
            ssl_context=ssl_context,
            connections_max_idle_ms=connections_max_idle_ms,
            sasl_mechanism=sasl_mechanism,
            sasl_plain_username=sasl_plain_username,
            sasl_plain_password=sasl_plain_password,
            sasl_kerberos_service_name=sasl_kerberos_service_name,
            sasl_kerberos_domain_name=sasl_kerberos_domain_name,
            sasl_oauth_token_provider=sasl_oauth_token_provider,
        )
        self._metadata = self.client.cluster
        self._closed = False

    # Warn if producer was not closed properly
    # We don't attempt to close the Consumer, as __del__ is synchronous
    def __del__(self, _warnings=warnings):
        if self._closed is False:
            _warnings.warn(
                f"Unclosed AIOKafkaProducer {self!r}",
                ResourceWarning,
                source=self,
            )
            context = {
                "producer": self,
                "message": "Unclosed AIOKafkaProducer",
            }
            if self._source_traceback is not None:
                context["source_traceback"] = self._source_traceback
            self._loop.call_exception_handler(context)

    @abc.abstractmethod
    def _on_set_api_version(self, api_version):
        ...

    @abc.abstractmethod
    def _message_accumulator_for(self, transactional_id, tp):
        ...

    @abc.abstractmethod
    def _transactional_id_or_default(self, transactional_id):
        ...

    @abc.abstractmethod
    def _verify_txn_started(self, transactional_id):
        ...

    @abc.abstractmethod
    def _wait_for_sender(self):
        ...

    @abc.abstractmethod
    def _ensure_transactional(self):
        ...

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, type, value, traceback):
        await self.stop()

    async def start(self):
        """Connect to Kafka cluster and check server version"""
        self.log.debug("Starting the Kafka producer")  # trace
        await self.client.bootstrap()
        if self._closed:
            return
        api_version = self.client.api_version

        self._verify_api_version(api_version)
        await self._start_sender()
        self._on_set_api_version(api_version)

        self._producer_magic = 0 if api_version < (0, 10) else 1
        self.log.debug("Kafka producer started")

    def _verify_api_version(self, api_version):
        if self._compression_type == "lz4":
            assert self.client.api_version >= (0, 8, 2), (
                "LZ4 Requires >= Kafka 0.8.2 Brokers"
            )  # fmt: skip
        elif self._compression_type == "zstd":
            assert self.client.api_version >= (2, 1, 0), (
                "Zstd Requires >= Kafka 2.1.0 Brokers"
            )  # fmt: skip

    async def stop(self):
        """Flush all pending data and close all connections to kafka cluster"""
        if self._closed:
            return
        self._closed = True

        await self._wait_for_sender()

        await self.client.close()
        self.log.debug("The Kafka producer has closed.")

    async def partitions_for(self, topic):
        """Returns set of all known partitions for the topic."""
        return await self.client._wait_on_metadata(topic)

    def _serialize(self, topic, key, value):
        if self._key_serializer:
            serialized_key = self._key_serializer(key)
        else:
            serialized_key = key
        if self._value_serializer:
            serialized_value = self._value_serializer(value)
        else:
            serialized_value = value

        message_size = LegacyRecordBatchBuilder.record_overhead(
            self._producer_magic)
        if serialized_key is not None:
            message_size += len(serialized_key)
        if serialized_value is not None:
            message_size += len(serialized_value)
        if message_size > self._max_request_size:
            raise MessageSizeTooLargeError(
                "The message is %d bytes when serialized which is larger than"
                " the maximum request size you have configured with the"
                " max_request_size configuration" % message_size)

        return serialized_key, serialized_value

    def _partition(self, topic, partition, key, value,
                   serialized_key, serialized_value):
        if partition is not None:
            assert partition >= 0
            assert partition in self._metadata.partitions_for_topic(topic), \
                'Unrecognized partition'
            return partition

        all_partitions = list(self._metadata.partitions_for_topic(topic))
        available = list(self._metadata.available_partitions_for_topic(topic))
        return self._partitioner(
            serialized_key, all_partitions, available)

    async def send(
        self, topic, value=None, key=None, partition=None,
        timestamp_ms=None, headers=None, transactional_id=None
    ):
        """Publish a message to a topic.

        Arguments:
            topic (str): topic where the message will be published
            value (optional): message value. Must be type bytes, or be
                serializable to bytes via configured value_serializer. If value
                is None, key is required and message acts as a 'delete'.
                See kafka compaction documentation for more details:
                http://kafka.apache.org/documentation.html#compaction
                (compaction requires kafka >= 0.8.1)
            partition (int, optional): optionally specify a partition. If not
                set, the partition will be selected using the configured
                'partitioner'.
            key (optional): a key to associate with the message. Can be used to
                determine which partition to send the message to. If partition
                is None (and producer's partitioner config is left as default),
                then messages with the same key will be delivered to the same
                partition (but if key is None, partition is chosen randomly).
                Must be type bytes, or be serializable to bytes via configured
                key_serializer.
            timestamp_ms (int, optional): epoch milliseconds (from Jan 1 1970
                UTC) to use as the message timestamp. Defaults to current time.

        Returns:
            asyncio.Future: object that will be set when message is
            processed

        Raises:
            kafka.KafkaTimeoutError: if we can't schedule this record (
                pending buffer is full) in up to `request_timeout_ms`
                milliseconds.

        Note:
            The returned future will wait based on `request_timeout_ms`
            setting. Cancelling the returned future **will not** stop event
            from being sent, but cancelling the ``send`` coroutine itself
            **will**.
        """
        assert value is not None or self.client.api_version >= (0, 8, 1), (
            'Null messages require kafka >= 0.8.1')
        assert not (value is None and key is None), \
            'Need at least one: key or value'
        transactional_id = self._transactional_id_or_default(transactional_id)

        # first make sure the metadata for the topic is available
        await self.client._wait_on_metadata(topic)

        # Ensure transaction is started and not committing
        self._verify_txn_started(transactional_id)

        if headers is not None:
            if self.client.api_version < (0, 11):
                raise UnsupportedVersionError(
                    "Headers not supported before Kafka 0.11")
        else:
            # Record parser/builder support only list type, no explicit None
            headers = []

        key_bytes, value_bytes = self._serialize(topic, key, value)
        partition = self._partition(topic, partition, key, value,
                                    key_bytes, value_bytes)

        tp = TopicPartition(topic, partition)
        self.log.debug("Sending (key=%s value=%s) to %s", key, value, tp)

        message_accumulator = self._message_accumulator_for(
            transactional_id, tp)
        fut = await message_accumulator.add_message(
            tp, key_bytes, value_bytes, self._request_timeout_ms / 1000,
            timestamp_ms=timestamp_ms, headers=headers)
        return fut

    async def send_and_wait(self, topic, value=None, key=None, partition=None,
                      timestamp_ms=None):
        """Publish a message to a topic and wait the result"""
        future = await self.send(
            topic, value, key, partition, timestamp_ms)
        return (await future)
    
class MultiTXNProducer(BaseProducer):
    """Kafka producer using :pypi:`aiokafka` with support for multi-transactions."""

    _transactions: MutableMapping[str, TransactionManager]
    _accumulators: MutableMapping[str, MessageAccumulator]
    _senders: MutableMapping[str, Sender]
    _received_api_version: str = None

    def __init__(self, *,
                 loop,
                 bootstrap_servers='localhost',
                 acks=_missing,
                 enable_idempotence=False,
                 transactional_id=None,
                 transaction_timeout_ms=60000,
                 **kwargs):
        if acks is _missing:
            acks = -1
        elif acks not in ('all', -1):
            raise ValueError(
                "acks={} not supported with MultiTXNProducer"
                .format(acks))

        super().__init__(
            loop=loop,
            bootstrap_servers=bootstrap_servers,
            acks=acks,
            transaction_timeout_ms=transaction_timeout_ms,
            **kwargs,
        )
        self._transactions = {}
        self._accumulators = {}
        self._senders = {}

        self._message_accumulator = MessageAccumulator(
            self._metadata, 
            self._max_batch_size, 
            self._compression_attrs,
            self._request_timeout_ms / 1000,
            txn_manager=None,
            loop=self._loop)

        self._sender = Sender(
            self.client,
            txn_manager=None,
            acks=self._acks,
            retry_backoff_ms=self._retry_backoff_ms,
            linger_ms=self._linger_ms,
            message_accumulator=self._message_accumulator,
            request_timeout_ms=self._request_timeout_ms,
        )

    async def commit(
        self,
        tid_to_offset_map: Mapping[str, Mapping[TopicPartition, int]],
        group_id: str,
        start_new_transaction: bool = True
    ) -> None:
        for transactional_id, offsets in tid_to_offset_map.items():
            self.log.debug('+COMMIT %r %r' % (transactional_id, offsets))
            try:
                await self._commit(
                    transactional_id, offsets, group_id,
                    start_new_transaction=start_new_transaction,
                )
                self.log.debug('-COMMIT %r %r' % (transactional_id, offsets))
            except Exception as exc:
                self.log.error(
                    f"Failed to commit transaction {transactional_id}: {exc}")
                raise

    async def _commit(
        self, transactional_id, offsets: Mapping[TopicPartition, int],
        group_id: str,
        start_new_transaction: bool = True
    ) -> None:
        self.log.debug('+send offsets to transaction %r' % (transactional_id,))
        await self.send_offsets_to_transaction(
            transactional_id, offsets, group_id)
        self.log.debug('-send offsets to transaction %r' % (transactional_id,))
        self.log.debug('+commit transaction %r' % (transactional_id,))
        await self.commit_transaction(transactional_id)
        self.log.debug('-commit transaction %r' % (transactional_id,))
        if start_new_transaction:
            self.log.debug('+start new transaction %r' % (transactional_id,))
            await self.begin_transaction(transactional_id)
            self.log.debug('-start new transaction %r' % (transactional_id,))

    def _on_set_api_version(self, api_version):
        self._received_api_version = api_version
        for accumulator in self._accumulators.values():
            accumulator.set_api_version(api_version)
        self._message_accumulator.set_api_version(api_version)

    def _message_accumulator_for(self, transactional_id, tp):
        if transactional_id is None:
            return self._message_accumulator
        return self._accumulators[transactional_id]

    def _transactional_id_or_default(self, transactional_id):
        return None

    def _verify_txn_started(self, transactional_id):
        try:
            txn_manager = self._transactions[transactional_id]
        except KeyError:
            pass
        else:
            assert txn_manager.transactional_id == transactional_id
            if not txn_manager.is_in_transaction():
                raise IllegalOperation(
                    "Can't send messages while not in transaction")

    def _verify_api_version(self, api_version):
        super()._verify_api_version(api_version)
        if self.client.api_version < (0, 11):
            raise UnsupportedVersionError(
                "MultiTXNProducer available only for Broker version 0.11"
                " and above")

    async def flush(self):
        """Wait untill all batches are Delivered and futures resolved"""
        await asyncio.gather(
            self._message_accumulator.flush(),
            *[acc.flush() for acc in self._accumulators.values()],
        )

    async def _wait_for_sender(self):
        senders = self._senders
        accumulators = self._accumulators
        await asyncio.gather(
            self._wait_for_sender1(self._sender, self._message_accumulator),
            *[self._wait_for_sender1(senders[transactional_id], accumulator)
              for transactional_id, accumulator in accumulators.items()],
        )

    async def _wait_for_sender1(self, sender, accumulator):
        # If the sender task is down there is no way for accumulator to flush
        if sender is not None and sender.sender_task is not None:
            await asyncio.wait(
                [
                    create_task(accumulator.close()),
                    sender.sender_task,
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )

            await sender.close()

    def _ensure_transactional(self):
        ...

    async def _start_sender(self):
        await self._sender.start()

    async def _init_transaction(self, tid):
        txn_manager = self._transactions[tid] = TransactionManager(
            tid, self._transaction_timeout_ms,
        )
        accumulator = self._accumulators[tid] = MessageAccumulator(
            self._metadata,
            self._max_batch_size,
            self._compression_attrs,
            self._request_timeout_ms / 1000,
            txn_manager=txn_manager,
            loop=self._loop,
        )
        accumulator.set_api_version(self._received_api_version)
        sender = self._senders[tid] = Sender(
            self.client,
            acks=self._acks,
            txn_manager=txn_manager,
            retry_backoff_ms=self._retry_backoff_ms,
            linger_ms=self._linger_ms,
            message_accumulator=accumulator,
            request_timeout_ms=self._request_timeout_ms,
        )
        await sender.start()
        return txn_manager

    async def begin_transaction(self, transactional_id):
        self.log.debug(
            "Beginning a new transaction for id %s", transactional_id)
        txn_manager = self._transactions.get(transactional_id)
        if txn_manager is None:
            txn_manager = await self._init_transaction(transactional_id)

        await asyncio.shield(
            txn_manager.wait_for_pid(),
            loop=self._loop,
        )
        txn_manager.begin_transaction()

    async def commit_transaction(self, transactional_id):
        self.log.debug(
            "Committing transaction for id %s", transactional_id)
        txn_manager = self._transactions[transactional_id]
        txn_manager.committing_transaction()
        await asyncio.shield(
            txn_manager.wait_for_transaction_end(),
            loop=self._loop,
        )

    async def abort_transaction(self, transactional_id):
        self.log.debug(
            "Aborting transaction for id %s", transactional_id)
        txn_manager = self._transactions[transactional_id]
        txn_manager.aborting_transaction()
        await asyncio.shield(
            txn_manager.wait_for_transaction_end(),
            loop=self._loop,
        )

    async def stop_transaction(self, transactional_id):
        txn_manager = self._transactions.pop(transactional_id, None)
        accumulator = self._accumulators.pop(transactional_id, None)
        sender = self._senders.pop(transactional_id, None)
        if txn_manager is not None:
            if txn_manager.is_in_transaction():
                txn_manager.aborting_transaction()
                await asyncio.shield(
                    txn_manager.wait_for_transaction_end(),
                    loop=self._loop,
                )
        await self._wait_for_sender1(sender, accumulator)

    async def maybe_begin_transaction(self, transactional_id):
        txn_manager = self._transactions.get(transactional_id)
        if txn_manager is None:
            txn_manager = await self._init_transaction(transactional_id)
        else:
            if txn_manager.is_in_transaction():
                return
        self.log.debug(
            "Beginning a new transaction for id %s", transactional_id)
        await asyncio.shield(
            txn_manager.wait_for_pid(),
            loop=self._loop,
        )
        txn_manager.begin_transaction()

    async def send_offsets_to_transaction(
        self, transactional_id, offsets, group_id
    ):
        txn_manager = self._transactions[transactional_id]
        if not txn_manager.is_in_transaction():
            raise IllegalOperation("Not in the middle of a transaction")

        if not group_id or not isinstance(group_id, str):
            raise ValueError(group_id)

        # validate `offsets` structure
        formatted_offsets = commit_structure_validate(offsets)

        self.log.debug(
            "Begin adding offsets %s for consumer group %s to transaction",
            formatted_offsets, group_id)
        fut = txn_manager.add_offsets_to_txn(formatted_offsets, group_id)
        self.log.debug('+WAIT FOR RESPONSE OR ERROR %r' % (fut,))
        await asyncio.shield(fut, loop=self._loop)
        self.log.debug('-WAIT FOR RESPONSE OR ERROR %r' % (fut,))


class Producer(base.Producer):
    """Kafka producer using :pypi:`aiokafka`."""

    logger = logger

    allow_headers: bool = True
    _producer: Optional[aiokafka.AIOKafkaProducer] = None

    def __post_init__(self) -> None:
        self._send_on_produce_message = self.app.on_produce_message.send
        if self.partitioner is None:
            self.partitioner = DefaultPartitioner()
        if self._api_version != 'auto':
            wanted_api_version = parse_kafka_version(self._api_version)
            if wanted_api_version < (0, 11):
                self.allow_headers = False

    def _settings_default(self) -> Mapping[str, Any]:
        transport = cast(Transport, self.transport)
        return {
            'bootstrap_servers': server_list(
                transport.url, transport.default_port),
            'client_id': self.client_id,
            'acks': self.acks,
            'linger_ms': self.linger_ms,
            'max_batch_size': self.max_batch_size,
            'max_request_size': self.max_request_size,
            'compression_type': self.compression_type,
            'security_protocol': 'SSL' if self.ssl_context else 'PLAINTEXT',
            'partitioner': self.partitioner,
            'request_timeout_ms': int(self.request_timeout * 1000),
            'api_version': self._api_version,
        }

    def _settings_auth(self) -> Mapping[str, Any]:
        return credentials_to_aiokafka_auth(
            self.credentials, self.ssl_context)

    async def begin_transaction(self, transactional_id: str) -> None:
        """Begin transaction by id."""
        await self._ensure_producer().begin_transaction(transactional_id)

    async def commit_transaction(self, transactional_id: str) -> None:
        """Commit transaction by id."""
        await self._ensure_producer().commit_transaction(transactional_id)

    async def abort_transaction(self, transactional_id: str) -> None:
        """Abort and rollback transaction by id."""
        await self._ensure_producer().abort_transaction(transactional_id)

    async def stop_transaction(self, transactional_id: str) -> None:
        """Stop transaction by id."""
        await self._ensure_producer().stop_transaction(transactional_id)

    async def maybe_begin_transaction(self, transactional_id: str) -> None:
        """Begin transaction (if one does not already exist)."""
        await self._ensure_producer().maybe_begin_transaction(transactional_id)

    async def commit_transactions(
            self,
            tid_to_offset_map: Mapping[str, Mapping[TP, int]],
            group_id: str,
            start_new_transaction: bool = True) -> None:
        """Commit transactions."""
        await self._ensure_producer().commit(
            tid_to_offset_map, group_id,
            start_new_transaction=start_new_transaction,
        )

    def _settings_extra(self) -> Mapping[str, Any]:
        if self.app.in_transaction:
            return {'acks': 'all'}
        return {}

    def _new_producer(self) -> aiokafka.AIOKafkaProducer:
        return self._producer_type(
            loop=self.loop,
            **{**self._settings_default(),
               **self._settings_auth(),
               **self._settings_extra()},
        )

    @property
    def _producer_type(self) -> Type[BaseProducer]:
        if self.app.in_transaction:
            return MultiTXNProducer
        return aiokafka.AIOKafkaProducer

    async def create_topic(self,
                           topic: str,
                           partitions: int,
                           replication: int,
                           *,
                           config: Mapping[str, Any] = None,
                           timeout: Seconds = 20.0,
                           retention: Seconds = None,
                           compacting: bool = None,
                           deleting: bool = None,
                           ensure_created: bool = False) -> None:
        """Create/declare topic on server."""
        _retention = (int(want_seconds(retention) * 1000.0)
                      if retention else None)
        producer = self._ensure_producer()
        await cast(Transport, self.transport)._create_topic(
            self,
            producer.client,
            topic,
            partitions,
            replication,
            config=config,
            timeout=int(want_seconds(timeout) * 1000.0),
            retention=_retention,
            compacting=compacting,
            deleting=deleting,
            ensure_created=ensure_created,
        )
        await producer.client.force_metadata_update()  # Fixes #499

    def _ensure_producer(self) -> BaseProducer:
        if self._producer is None:
            raise NotReady('Producer service not yet started')
        return self._producer

    async def on_start(self) -> None:
        """Call when producer starts."""
        await super().on_start()
        producer = self._producer = self._new_producer()
        self.beacon.add(producer)
        await producer.start()

    async def on_stop(self) -> None:
        """Call when producer stops."""
        await super().on_stop()
        cast(Transport, self.transport)._topic_waiters.clear()
        producer, self._producer = self._producer, None
        if producer is not None:
            await producer.stop()

    async def send(self, topic: str, key: Optional[bytes],
                   value: Optional[bytes],
                   partition: Optional[int],
                   timestamp: Optional[float],
                   headers: Optional[HeadersArg],
                   *,
                   transactional_id: str = None) -> Awaitable[RecordMetadata]:
        """Schedule message to be transmitted by producer."""
        producer = self._ensure_producer()
        if headers is not None:
            if isinstance(headers, Mapping):
                headers = list(headers.items())
        self._send_on_produce_message(
            key=key, value=value,
            partition=partition,
            timestamp=timestamp,
            headers=headers,
        )
        if headers is not None and not self.allow_headers:
            headers = None
        timestamp_ms = int(timestamp * 1000.0) if timestamp else timestamp
        try:
            return cast(Awaitable[RecordMetadata], await producer.send(
                topic, value,
                key=key,
                partition=partition,
                timestamp_ms=timestamp_ms,
                headers=headers,
                transactional_id=transactional_id,
            ))
        except KafkaError as exc:
            raise ProducerSendError(f'Error while sending: {exc!r}') from exc

    async def send_and_wait(self, topic: str, key: Optional[bytes],
                            value: Optional[bytes],
                            partition: Optional[int],
                            timestamp: Optional[float],
                            headers: Optional[HeadersArg],
                            *,
                            transactional_id: str = None) -> RecordMetadata:
        """Send message and wait for it to be transmitted."""
        fut = await self.send(
            topic,
            key=key,
            value=value,
            partition=partition,
            timestamp=timestamp,
            headers=headers,
            transactional_id=transactional_id,
        )
        return await fut

    async def flush(self) -> None:
        """Wait for producer to finish transmitting all buffered messages."""
        await self.buffer.flush()
        if self._producer is not None:
            await self._producer.flush()

    def key_partition(self, topic: str, key: bytes) -> TP:
        """Hash key to determine partition destination."""
        producer = self._ensure_producer()
        partition = producer._partition(
            topic,
            partition=None,
            key=None,
            value=None,
            serialized_key=key,
            serialized_value=None,
        )
        return TP(topic, partition)

    def supports_headers(self) -> bool:
        """Return :const:`True` if message headers are supported."""
        producer = self._ensure_producer()
        client = producer.client
        if client is None:
            raise NotReady('Producer client not yet connected')
        return client.api_version >= (0, 11)


class Transport(base.Transport):
    """Kafka transport using :pypi:`aiokafka`."""

    Consumer: ClassVar[Type[ConsumerT]]
    Consumer = Consumer
    Producer: ClassVar[Type[ProducerT]]
    Producer = Producer

    default_port = 9092
    driver_version = f'aiokafka={aiokafka.__version__}'

    _topic_waiters: MutableMapping[str, StampedeWrapper]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._topic_waiters = {}

    def _topic_config(self,
                      retention: int = None,
                      compacting: bool = None,
                      deleting: bool = None) -> MutableMapping[str, Any]:
        config: MutableMapping[str, Any] = {}
        cleanup_flags: Set[str] = set()
        if compacting:
            cleanup_flags |= {'compact'}
        if deleting:
            cleanup_flags |= {'delete'}
        if cleanup_flags:
            config['cleanup.policy'] = ','.join(sorted(cleanup_flags))
        if retention:
            config['retention.ms'] = retention
        return config

    async def _create_topic(self,
                            owner: Service,
                            client: aiokafka.AIOKafkaClient,
                            topic: str,
                            partitions: int,
                            replication: int,
                            **kwargs: Any) -> None:
        assert topic is not None
        try:
            wrap = self._topic_waiters[topic]
        except KeyError:
            wrap = self._topic_waiters[topic] = StampedeWrapper(
                self._really_create_topic,
                owner,
                client,
                topic,
                partitions,
                replication,
                loop=asyncio.get_event_loop(), **kwargs)
        try:
            await wrap()
        except Exception:
            self._topic_waiters.pop(topic, None)
            raise

    async def _get_controller_node(
            self,
            owner: Service,
            client: aiokafka.AIOKafkaClient,
            timeout: int = 30000) -> Optional[int]:  # pragma: no cover
        nodes = [broker.nodeId for broker in client.cluster.brokers()]
        for node_id in nodes:
            if node_id is None:
                raise NotReady('Not connected to Kafka Broker')
            request = MetadataRequest_v1([])
            wait_result = await owner.wait(
                client.send(node_id, request),
                timeout=timeout,
            )
            if wait_result.stopped:
                owner.log.info('Shutting down - skipping creation.')
                return None
            response = wait_result.result
            return response.controller_id
        raise Exception('Controller node not found')

    async def _really_create_topic(
            self,
            owner: Service,
            client: aiokafka.AIOKafkaClient,
            topic: str,
            partitions: int,
            replication: int,
            *,
            config: Mapping[str, Any] = None,
            timeout: int = 30000,
            retention: int = None,
            compacting: bool = None,
            deleting: bool = None,
            ensure_created: bool = False) -> None:  # pragma: no cover
        owner.log.info('Creating topic %r', topic)

        if topic in client.cluster.topics():
            owner.log.debug('Topic %r exists, skipping creation.', topic)
            return

        protocol_version = 1
        extra_configs = config or {}
        config = self._topic_config(retention, compacting, deleting)
        config.update(extra_configs)

        controller_node = await self._get_controller_node(owner, client,
                                                          timeout=timeout)
        owner.log.debug('Found controller: %r', controller_node)

        if controller_node is None:
            if owner.should_stop:
                owner.log.info('Shutting down hence controller not found')
                return
            else:
                raise Exception('Controller node is None')

        request = CreateTopicsRequest[protocol_version](
            [(topic, partitions, replication, [], list(config.items()))],
            timeout,
            False,
        )
        wait_result = await owner.wait(
            client.send(controller_node, request),
            timeout=timeout,
        )
        if wait_result.stopped:
            owner.log.debug('Shutting down - skipping creation.')
            return
        response = wait_result.result

        assert len(response.topic_errors), "single topic"

        _, code, reason = response.topic_errors[0]

        if code != 0:
            if not ensure_created and code == TopicExistsError.errno:
                owner.log.debug("Topic %r exists, skipping creation.", topic)
                return
            elif code == NotControllerError.errno:
                raise RuntimeError(f"Invalid controller: {controller_node}")
            else:
                raise for_code(code)(f"Cannot create topic: {topic} ({code}): {reason}")
        else:
            owner.log.info("Topic %r created.", topic)
            return


def credentials_to_aiokafka_auth(credentials: CredentialsT = None,
                                 ssl_context: Any = None) -> Mapping:
    if credentials is not None:
        if isinstance(credentials, SSLCredentials):
            return {
                'security_protocol': credentials.protocol.value,
                'ssl_context': credentials.context,
            }
        elif isinstance(credentials, SASLCredentials):
            return {
                'security_protocol': credentials.protocol.value,
                'sasl_mechanism': credentials.mechanism.value,
                'sasl_plain_username': credentials.username,
                'sasl_plain_password': credentials.password,
                'ssl_context': credentials.ssl_context,
            }
        elif isinstance(credentials, GSSAPICredentials):
            return {
                'security_protocol': credentials.protocol.value,
                'sasl_mechanism': credentials.mechanism.value,
                'sasl_kerberos_service_name':
                    credentials.kerberos_service_name,
                'sasl_kerberos_domain_name':
                    credentials.kerberos_domain_name,
                'ssl_context': credentials.ssl_context,
            }
        else:
            raise ImproperlyConfigured( 
                f'aiokafka does not support {credentials}')
    elif ssl_context is not None:
        return {
            'security_protocol': 'SSL',
            'ssl_context': ssl_context,
        }
    else:
        return {'security_protocol': 'PLAINTEXT'}



import asyncio
import pytest
from mode.utils.mocks import ANY, AsyncMock, Mock, patch
from faust.livecheck.exceptions import (
    LiveCheckError,
    TestFailed,
    TestRaised,
    TestSkipped,
    TestTimeout,
)
from faust.livecheck.models import State


class test_TestRunner:

    @pytest.mark.asyncio
    async def test_execute__case_inactive(self, *, runner, execution):
        with pytest.raises(TestSkipped):
            await self._do_execute(runner, execution, active=False)

    @pytest.mark.asyncio
    async def test_execute__test_expired(self, *, runner, execution):
        with pytest.raises(TestSkipped):
            await self._do_execute(runner, execution, expired=True)

    @pytest.mark.asyncio
    @pytest.mark.parametrize('exc,raises,callback', [
        (TestSkipped('foo'), None, 'on_skipped'),
        (TestTimeout('bar'), None, 'on_timeout'),
        (AssertionError('baz'), TestFailed, 'on_failed'),
        (LiveCheckError('xuz'), None, 'on_error'),
        (KeyError('muz'), TestRaised, 'on_error'),
    ])
    async def test_execute__error_callbacks(self, exc, raises, callback, *,
                                            runner, execution):
        if raises is None:
            raises = type(exc)

        cb = AsyncMock()
        setattr(runner, callback, cb)
        runner.on_pass = AsyncMock()

        with pytest.raises(raises):
            await self._do_execute(
                runner, execution,
                active=True,
                expired=False,
                side_effect=exc,
            )
        cb.assert_called_once_with(exc)
        runner.on_pass.assert_not_called()

    @pytest.mark.asyncio
    async def test_execute__pass(self, *, runner, execution):
        runner.on_pass = AsyncMock()
        await self._do_execute(runner, execution)
        runner.on_pass.coro.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_execute__CancelledError(self, *, runner, execution):
        await self._do_execute(runner, execution,
                               side_effect=asyncio.CancelledError())

    async def _do_execute(self, runner, execution,
                          active=True, expired=False, side_effect=None):
        runner.case.active = active
        runner.test = execution
        runner.test.is_expired = expired
        runner.case.run = AsyncMock()
        if side_effect:
            runner.case.run.side_effect = side_effect
        await runner.execute()

    @pytest.mark.asyncio
    async def test_skip(self, runner):
        with pytest.raises(TestSkipped):
            runner.on_skipped = AsyncMock()
            await runner.skip('broken')
            runner.on_skipped.coro.assert_called_once_with(ANY)

    def test__prepare_args(self, *, runner):
        assert runner._prepare_args((1, 2, 3, object()))

    def test__prepare_kwargs(self, *, runner):
        assert runner._prepare_kwargs({
            'foo': object(),
            'bar': 1,
            'baz': 1.03,
        })

    @pytest.mark.asyncio
    async def test_on_skipped(self, *, runner):
        runner.case.on_test_skipped = AsyncMock()
        runner.state = State.PASS
        exc = TestSkipped()
        await runner.on_skipped(exc)
        assert runner.state == State.SKIP
        runner.case.on_test_skipped.assert_called_once_with(runner)

    @pytest.mark.asyncio
    async def test_on_start(self, *, runner):
        runner.case.on_test_start = AsyncMock()
        await runner.on_start()
        runner.case.on_test_start.assert_called_once_with(runner)

    @pytest.mark.asyncio
    async def test_on_signal_wait(self, *, runner):
        signal = Mock()
        signal.name = 'foo'
        signal.index = 2
        await runner.on_signal_wait(signal, 30.0)

    @pytest.mark.asyncio
    async def test_on_signal_received(self, *, runner):
        signal = Mock()
        signal.name = 'order_in_db'
        await runner.on_signal_received(signal, 100.0, 200.0)
        assert runner.signal_latency[signal.name] == 100.0

    @pytest.mark.asyncio
    async def test_on_failed(self, *, runner):
        runner.error = KeyError()
        runner.case.on_test_failed = AsyncMock()
        runner._finalize_report = AsyncMock()

        exc = TestFailed('foo the bar')
        await runner.on_failed(exc)

        assert runner.error is exc
        assert runner.state == State.FAIL
        runner.case.on_test_failed.coro.assert_called_once_with(runner, exc)
        runner._finalize_report.coro.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_error(self, *, runner):
        runner.error = KeyError()
        runner.case.on_test_error = AsyncMock()
        runner._finalize_report = AsyncMock()

        exc = TestRaised('foo the bar')
        await runner.on_error(exc)

        assert runner.error is exc
        assert runner.state == State.ERROR
        runner.case.on_test_error.coro.assert_called_once_with(runner, exc)
        runner._finalize_report.coro.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_timeout(self, *, runner):
        runner.error = KeyError()
        runner.case.on_test_timeout = AsyncMock()
        runner._finalize_report = AsyncMock()

        exc = TimeoutError('foo the bar')
        await runner.on_timeout(exc)

        assert runner.error is exc
        assert runner.state == State.TIMEOUT
        runner.case.on_test_timeout.coro.assert_called_once_with(runner, exc)
        runner._finalize_report.coro.assert_called_once_with()

    @pytest.mark.asyncio
    async def test_on_pass(self, *, runner):
        runner.error = KeyError()
        runner.case.on_test_pass = AsyncMock()
        runner._finalize_report = AsyncMock()
        await runner.on_pass()

        assert runner.state == State.PASS
        assert runner.error is None

        runner.case.on_test_pass.coro.assert_called_once_with(runner)
        runner._finalize_report.coro.assert_called_once_with()

    @pytest.mark.asyncio
    async def test__finalize_report(self, *, runner):
        exc = None
        try:
            raise KeyError('foo')
        except KeyError as e:
            exc = e
        runner.error = exc
        runner.case.post_report = AsyncMock()
        runner.state = State.ERROR
        runner.runtime = 3.011
        await runner._finalize_report()

        report = runner.report
        assert report.case_name == runner.case.name
        assert report.state == State.ERROR
        assert report.test == runner.test
        assert report.runtime == runner.runtime
        assert report.signal_latency == {}
        assert report.error == str(exc)
        assert report.traceback
        runner.case.post_report.coro.assert_called_once_with(report)

    @pytest.mark.asyncio
    async def test__finalize_report__no_error(self, *, runner):
        runner.error = None
        runner.case.post_report = AsyncMock()
        runner.state = State.ERROR
        runner.runtime = 3.011
        await runner._finalize_report()

        report = runner.report
        assert report.case_name == runner.case.name
        assert report.state == State.ERROR
        assert report.test == runner.test
        assert report.runtime == runner.runtime
        assert report.signal_latency == {}
        assert report.error is None
        assert report.traceback is None
        runner.case.post_report.coro.assert_called_once_with(report)

    def test_log_info(self, *, runner):
        runner.case.realtime_logs = False
        runner.log_info('msg %r', 1)
        assert runner.logs == [('msg %r', (1,))]

    def test_log_info__realtime(self, *, runner):
        runner.case.realtime_logs = True
        runner.log = Mock()
        runner.log_info('msg %r', 1)
        runner.log.info.assert_called_once_with('msg %r', 1)

    def test_end(self, *, runner):
        with patch('faust.livecheck.runners.monotonic') as monotonic:
            monotonic.return_value = 100.0
            runner.started = 50.0
            runner.end()

            assert runner.ended == 100.0
            assert runner.runtime == 50.0

    def test__flush_logs(self, *, runner):
        runner.logs = [
            ('foo %r bar', (10,)),
            ('The test %r raised %r', ('foo', 'error')),
        ]
        runner.log = Mock()
        runner._flush_logs()
        assert runner.log.logger.log.call_count == 1
        assert 'The test' in runner.log.logger.log.call_args[0][1]
        assert 'foo 10 bar' in runner.log.logger.log.call_args[0][1]
        assert not runner.logs



"""Agent replies: waiting for replies, sending them, etc."""
import asyncio

from collections import defaultdict
from typing import (
    Any,
    AsyncIterator,
    MutableMapping,
    MutableSet,
    NamedTuple,
    Optional,
)
from weakref import WeakSet

from mode import Service

from faust.types import AppT, ChannelT, TopicT
from .models import ReqRepResponse

__all__ = ['ReplyPromise', 'BarrierState', 'ReplyConsumer']


class ReplyTuple(NamedTuple):
    correlation_id: str
    value: Any


class ReplyPromise(asyncio.Future):
    """Reply promise can be :keyword:`await`-ed to wait until result ready."""

    reply_to: str
    correlation_id: str

    def __init__(self, reply_to: str, correlation_id: str = '',
                 **kwargs: Any) -> None:
        self.reply_to = reply_to
        self._verify_correlation_id(correlation_id)
        self.correlation_id = correlation_id
        self.__post_init__()
        super().__init__(**kwargs)

    def __post_init__(self) -> None:
        ...

    def _verify_correlation_id(self, correlation_id: str) -> None:
        if not correlation_id:
            raise ValueError('ReplyPromise missing correlation_id argument.')

    def fulfill(self, correlation_id: str, value: Any) -> None:
        """Fulfill promise: a reply was received."""
        # If it wasn't for BarrierState we would just use .set_result()
        # directly, but BarrierState.fulfill requires the correlation_id
        # to be sent with it. That way it can mark that part of the map
        # operation as completed.
        assert correlation_id == self.correlation_id
        self.set_result(value)


class BarrierState(ReplyPromise):
    """State of pending/complete barrier.

    A barrier is a synchronization primitive that will wait until
    a group of coroutines have completed.
    """

    #: This is the size while the messages are being sent.
    #: (it's a tentative total, added to until the total is finalized).
    size: int = 0

    #: This is the actual total when all messages have been sent.
    #: It's set by :meth:`finalize`.
    total: int = 0

    #: The number of results we have received.
    fulfilled: int = 0

    #: Internal queue where results are added to.
    _results: asyncio.Queue

    #: Set of pending replies that this barrier is composed of.
    pending: MutableSet[ReplyPromise]

    def __post_init__(self) -> None:
        self.pending = set()
        loop: asyncio.AbstractEventLoop = self._loop
        self._results = asyncio.Queue(maxsize=1000, loop=loop)

    def _verify_correlation_id(self, correlation_id: str) -> None:
        pass  # barrier does not require a correlation id.

    def add(self, p: ReplyPromise) -> None:
        """Add promise to barrier.

        Note:
            You can only add promises before the barrier is finalized
            using :meth:`finalize`.
        """
        self.pending.add(p)
        self.size += 1

    def finalize(self) -> None:
        """Finalize this barrier.

        After finalization you can not grow or shrink the size
        of the barrier.
        """
        self.total = self.size
        # The barrier may have been filled up already at this point,
        if self.fulfilled >= self.total:
            self.set_result(True)
            self._results.put_nowait(None)  # always wake-up .iterate()

    def fulfill(self, correlation_id: str, value: Any) -> None:
        """Fulfill one of the promises in this barrier.

        Once all promises in this barrier is fulfilled, the barrier
        will be ready.
        """
        # ReplyConsumer calls this whenever a new reply is received.
        self._results.put_nowait(ReplyTuple(correlation_id, value))
        self.fulfilled += 1
        if self.total:
            if self.fulfilled >= self.total:
                self.set_result(True)
                self._results.put_nowait(None)  # always wake-up .iterate()

    def get_nowait(self) -> ReplyTuple:
        """Return next reply, or raise :exc:`asyncio.QueueEmpty`."""
        for _ in range(10):  # remove sentinels
            value = self._results.get_nowait()
            if value is not None:
                return value
        raise asyncio.QueueEmpty()

    async def iterate(self) -> AsyncIterator[ReplyTuple]:
        """Iterate over results as they arrive."""
        get = self._results.get
        get_nowait = self._results.get_nowait
        is_done = self.done
        while not is_done():
            value = await get()
            if value is not None:
                yield value
        while 1:
            try:
                value = get_nowait()
            except asyncio.QueueEmpty:
                break
            else:
                if value is not None:
                    yield value


class ReplyConsumer(Service):
    """Consumer responsible for redelegation of replies received."""

    _waiting: MutableMapping[str, MutableSet[ReplyPromise]]
    _fetchers: MutableMapping[str, Optional[asyncio.Future]]

    def __init__(self, app: AppT, **kwargs: Any) -> None:
        self.app = app
        self._waiting = defaultdict(WeakSet)
        self._fetchers = {}
        super().__init__(**kwargs)

    async def on_start(self) -> None:
        """Call when reply consumer starts."""
        if self.app.conf.reply_create_topic:
            await self._start_fetcher(self.app.conf.reply_to)

    async def add(self, correlation_id: str, promise: ReplyPromise) -> None:
        """Register promise to start tracking when it arrives."""
        reply_topic = promise.reply_to
        if reply_topic not in self._fetchers:
            await self._start_fetcher(reply_topic)
        self._waiting[correlation_id].add(promise)

    async def _start_fetcher(self, topic_name: str) -> None:
        if topic_name not in self._fetchers:
            # set the key as a lock, so it doesn't happen twice
            self._fetchers[topic_name] = None
            # declare the topic
            topic = self._reply_topic(topic_name)
            await topic.maybe_declare()
            await self.sleep(3.0)
            # then create the future
            self._fetchers[topic_name] = self.add_future(
                self._drain_replies(topic))

    async def _drain_replies(self, channel: ChannelT) -> None:
        async for reply in channel.stream():
            for promise in self._waiting[reply.correlation_id]:
                promise.fulfill(reply.correlation_id, reply.value)

    def _reply_topic(self, topic: str) -> TopicT:
        return self.app.topic(
            topic,
            partitions=1,
            replicas=0,
            deleting=True,
            retention=self.app.conf.reply_expires,
            value_type=ReqRepResponse,
        )



import asyncio
import os
import random
import subprocess
import sys
from t.consistency.consistency_checker import ConsistencyChecker


class Stresser(object):

    def __init__(self, num_workers, num_producers, loop):
        self.workers = set(range(num_workers))
        self._worker_procs = {}
        self.num_workers = num_workers
        self.num_producers = num_producers
        self.producers = set(range(num_producers))
        self._producer_procs = {}
        self.loop = loop
        self._stop_stresser = asyncio.Event(loop=loop)

    @property
    def _stopped(self):
        return self.workers - self._running

    @property
    def _running(self):
        return set(self._worker_procs)

    @property
    def _stopped_producers(self):
        return self.producers - self._running_producers

    @property
    def _running_producers(self):
        return set(self._producer_procs)

    async def _run_stresser(self):
        print('Running stresser')
        while not self._stop_stresser.is_set():
            print('Stresser iteration')
            if self._should():
                await self._maybe_stop_worker()
            if self._should():
                await self._maybe_spawn_worker()
            await asyncio.sleep(random.uniform(1, 20))

    def stop_stresser(self):
        print('Stopping stresser')
        self._stop_stresser.set()

    def _should(self):
        return (
            random.choices([True, False], [0.75, 0.25], k=1)[0] and
            not self._stop_stresser.is_set()
        )

    async def _maybe_stop_worker(self):
        print('Maybe stop')
        if len(self._running) > 1:
            await self._stop_worker(random.choice(list(self._running)))

    async def _maybe_spawn_worker(self):
        print('Maybe start')
        if self._stopped:
            await self._start_worker(random.choice(list(self._stopped)))

    async def start(self, stopped_at_start=0):
        assert stopped_at_start < self.num_workers
        start_workers = random.sample(list(self.workers),
                                      self.num_workers - stopped_at_start)
        print(f'Start workers: {start_workers!r}')
        await asyncio.wait([self._start_producer(producer)
                            for producer in self.producers])
        await asyncio.wait([self._start_worker(worker)
                            for worker in start_workers],
                           loop=self.loop, return_when=asyncio.ALL_COMPLETED)
        asyncio.ensure_future(self._run_stresser(), loop=self.loop)

    async def _start_worker(self, worker):
        assert worker in self.workers
        with open(f'worker_{worker}.logs', 'a') as f:
            if worker not in self._worker_procs:
                print(f'Starting worker {worker}')
                self._worker_procs[worker] = await self._exec_worker(
                    web_port=8080 + worker,
                    stdout=f,
                )

    async def _exec_worker(self,
                           module='faust',
                           app='examples.simple',
                           loglevel='info',
                           web_port=8080,
                           stdout=None,
                           stderr=subprocess.STDOUT,
                           **kwargs):
        return await asyncio.create_subprocess_exec(
            sys.executable, '-m', module,
            '-A', app,
            'worker',
            '-l', loglevel,
            '--web-port', str(web_port),
            stdout=stdout,
            stderr=stderr,
            env={**os.environ, **{'DEVLOG': '1'}},
            **kwargs)

    async def stop_all(self):
        await asyncio.wait(
            [self._stop_worker(worker) for worker in self._running],
            loop=self.loop, return_when=asyncio.ALL_COMPLETED,
        )

    async def stop_all_producers(self):
        await asyncio.wait(
            [self._stop_producer(producer)
             for producer in self._running_producers],
            loop=self.loop, return_when=asyncio.ALL_COMPLETED,
        )

    async def _stop_worker(self, worker):
        assert worker in self.workers
        print(f'Stopping worker {worker}')
        proc = self._worker_procs.pop(worker)
        await self._stop_process(proc)

    async def _start_producer(self, producer):
        assert producer in self.producers
        if producer not in self._producer_procs:
            with open(f'producer_{producer}.logs', 'a') as f:
                print(f'Starting producer: {producer}')
                self._producer_procs[producer] = await self._exec_producer(
                    stdout=f,
                )

    async def _exec_producer(self,
                             path='/Users/vineet/faust/examples/simple.py',
                             loglevel='info',
                             stdout=None,
                             stderr=subprocess.STDOUT,
                             **kwargs):
        return await asyncio.create_subprocess_exec(
            sys.executable,
            path,
            'produce',
            '-l', loglevel,
            stdout=stdout,
            stderr=stderr,
            **kwargs)

    async def _stop_producer(self, producer):
        assert producer in self.producers
        print(f'Stopping producer {producer}')
        proc = self._producer_procs.pop(producer)
        await self._stop_process(proc)

    async def _stop_process(self, proc):
        try:
            proc.terminate()
        except ProcessLookupError:
            pass
        await proc.wait()


async def test_consistency(loop):
    stresser = Stresser(num_workers=4, num_producers=4, loop=loop)
    checker = ConsistencyChecker('withdrawals',
                                 'f-simple-user_to_total-changelog', loop=loop)
    print('Starting stresser')
    await stresser.start(stopped_at_start=1)
    print('Waiting for stresser to run')
    await asyncio.sleep(180)  # seconds to run stresser for
    print('Stopping all producers')
    await stresser.stop_all_producers()
    await checker.build_source()
    print('Waiting to stop stresser')
    await asyncio.sleep(5.0)
    stresser.stop_stresser()
    await checker.wait_no_lag()
    print('Stopping everything')
    await stresser.stop_all()
    await checker.build_changelog()
    await checker.check_consistency()


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(test_consistency(loop))



import asyncio
import pytest
from faust.agents.models import ReqRepResponse
from faust.agents.replies import (
    BarrierState,
    ReplyConsumer,
    ReplyPromise,
)
from mode.utils.mocks import AsyncMock, Mock


def test_ReplyPromise():
    r = ReplyPromise(reply_to='rt', correlation_id='id1')
    assert r.reply_to == 'rt'
    assert r.correlation_id == 'id1'
    assert not r.done()
    r.fulfill('id1', 'value')
    assert r.result() == 'value'

    with pytest.raises(ValueError):
        r._verify_correlation_id('')
    with pytest.raises(ValueError):
        r._verify_correlation_id(None)


class test_BarrierState:

    @pytest.mark.asyncio
    async def test_parallel_join(self):
        p = BarrierState(reply_to='rt')
        assert not p.pending
        assert p.size == 0

        await asyncio.gather(
            self.adder(p),
            self.fulfiller(p),
            self.finalizer(p, 1.0),
            self.joiner(p),
        )

    @pytest.mark.asyncio
    async def test_get_nowait__exhaust_sentinels(self):
        p = BarrierState(reply_to='rt')
        for _ in range(20):
            p._results.put_nowait(None)
        with pytest.raises(asyncio.QueueEmpty):
            p.get_nowait()

    @pytest.mark.asyncio
    async def test_iterate__completion(self):
        p = BarrierState(reply_to='rt')
        p.done = Mock(name='done')
        p.done.return_value = False
        p._results.put_nowait(None)
        p._results.get = AsyncMock(name='get')

        def se():
            p.done.return_value = True
            return None

        p._results.get.coro.side_effect = se

        assert [x async for x in p.iterate()] == []

    @pytest.mark.asyncio
    async def test_parallel_iterate(self):
        p = BarrierState(reply_to='rt')
        assert not p.pending
        assert p.size == 0

        done, pending = await asyncio.wait(
            [
                self.adder(p),
                self.fulfiller(p),
                self.finalizer(p, 1.0),
                self.consumer(p),
            ],
            timeout=5.0,
        )

        if pending:
            raise Exception(
                f'Test did not return in 5s:\n'
                f'  DONE_TASKS={done}\n'
                f'  PENDING_TASKS={pending}\n'
                f'  size={p.size}\n'
                f'  total={p.total}\n'
                f'  fulfilled={p.fulfilled}\n'
                f'  pending={len(p.pending)}\n'
                f'  done={p.done()}'
                f'  result={p.result() if p.done() else None}')

    @pytest.mark.asyncio
    async def test_sync_join(self):
        p = BarrierState(reply_to='rt')
        assert not p.pending
        assert p.size == 0

        await self.adder(p)
        await self.fulfiller(p)
        await self.finalizer(p, 0.0)
        await self.joiner(p)
        assert p.get_nowait()

    @pytest.mark.asyncio
    async def test_sync_iterate(self):
        p = BarrierState(reply_to='rt')
        assert not p.pending
        assert p.size == 0

        await self.adder(p)
        await self.finalizer(p, 0.0)
        await self.fulfiller(p)
        assert p.done()
        await self.consumer(p)

    async def adder(self, p: BarrierState):
        for i in range(100):
            assert p.size == i
            r = ReplyPromise(reply_to=p.reply_to, correlation_id=str(i))
            p.add(r)
            assert len(p.pending) == i + 1
        await asyncio.sleep(0.5)
        assert p.size == i + 1

    async def finalizer(self, p: BarrierState, sleep_for):
        await asyncio.sleep(sleep_for)
        p.finalize()
        assert p.total == 100

    async def fulfiller(self, p: BarrierState):
        for i in range(100):
            await asyncio.sleep(0.01)
            p.fulfill(correlation_id=str(i), value=str(i))

    async def joiner(self, p: BarrierState):
        await p

    async def consumer(self, p: BarrierState):
        results = [None] * 100
        async for correlation_id, value in p.iterate():
            results[int(correlation_id)] = (correlation_id, value)
        for i, (correlation_id, value) in enumerate(results):
            assert correlation_id == str(i)
            assert value == str(i)


class test_ReplyConsumer:

    @pytest.fixture()
    def c(self, *, app):
        return ReplyConsumer(app)

    @pytest.mark.asyncio
    async def test_on_start__disabled(self, *, c, app):
        c._start_fetcher = AsyncMock()
        app.conf.reply_create_topic = False
        await c.on_start()
        c._start_fetcher.assert_not_called()

    @pytest.mark.asyncio
    async def test_on_start__enabled(self, *, c, app):
        c._start_fetcher = AsyncMock()
        app.conf.reply_create_topic = True
        await c.on_start()
        c._start_fetcher.assert_called_once_with(app.conf.reply_to)

    @pytest.mark.asyncio
    async def test_add(self, *, c):
        assert not c._waiting
        c._start_fetcher = AsyncMock()
        p = ReplyPromise(reply_to='rt', correlation_id='id1')
        await c.add('id1', p)
        assert 'id1' in c._waiting
        assert p in list(c._waiting['id1'])

        p2 = ReplyPromise(reply_to='rt', correlation_id='id1')
        c._fetchers[p2.reply_to] = Mock()
        await c.add('id1', p2)
        assert p in list(c._waiting['id1'])
        assert p2 in list(c._waiting['id1'])

        c._start_fetcher.assert_called_once_with(p.reply_to)

    @pytest.mark.asyncio
    async def test_start_fetcher(self, *, c):
        c._drain_replies = Mock()
        c._reply_topic = Mock(
            return_value=Mock(
                maybe_declare=AsyncMock(),
            ),
        )
        c.sleep = AsyncMock()
        c.add_future = Mock()

        await c._start_fetcher('topic1')
        assert 'topic1' in c._fetchers
        assert c._fetchers['topic1'] is c.add_future.return_value
        await c._start_fetcher('topic1')

        c._reply_topic.assert_called_once_with('topic1')
        topic = c._reply_topic.return_value
        topic.maybe_declare.assert_called_once_with()

        c.add_future.assert_called_once_with(
            c._drain_replies.return_value,
        )
        c._drain_replies.assert_called_once_with(topic)
        c.sleep.assert_called_once_with(3.0)

    @pytest.mark.asyncio
    async def test_drain_replies(self, *, c):
        responses = [
            ReqRepResponse(key='key1', value='value1', correlation_id='id1'),
            ReqRepResponse(key='key2', value='value2', correlation_id='id2'),
        ]
        channel = Mock(
            stream=Mock(return_value=self._response_stream(responses)),
        )
        p1 = Mock()
        p2 = Mock()
        p3 = Mock()
        c._waiting['id1'] = {p1, p2}
        c._waiting['id2'] = {p3}

        await c._drain_replies(channel)

        p1.fulfill.assert_called_once_with('id1', 'value1')
        p2.fulfill.assert_called_once_with('id1', 'value1')
        p3.fulfill.assert_called_once_with('id2', 'value2')

    async def _response_stream(self, responses):
        for response in responses:
            yield response

    def test_reply_topic(self, *, c, app):
        topic = c._reply_topic('foo')
        assert topic.get_topic_name() == 'foo'
        assert topic.partitions == 1
        assert topic.replicas == 0
        assert topic.deleting
        assert topic.retention == app.conf.reply_expires
        assert topic.value_type is ReqRepResponse



"""Worker.

A "worker" starts a single instance of a Faust application.

See Also:
    :ref:`app-starting`: for more information.
"""
import asyncio
import logging
import os
import sys
import psutil

from collections import defaultdict
from itertools import chain
from pathlib import Path
from typing import Any, Dict, IO, Iterable, Mapping, Optional, Set, Union

import mode
from aiokafka.structs import TopicPartition
from mode import ServiceT, get_logger
from mode.utils.logging import Severity, formatter2
from mode.utils.locks import Event

from .types import AppT, SensorT, TP, TopicT
from .types._env import (
    BLOCKING_TIMEOUT,
    CONSOLE_PORT,
    DEBUG,
    FORCE_BLOCKING_TIMEOUT,
)
from .utils import terminal
from .utils.functional import consecutive_numbers

try:  # pragma: no cover
    # if installed we use this to set ps.name (argv[0])
    from setproctitle import setproctitle
except ImportError:  # pragma: no cover
    def setproctitle(title: str) -> None: ...  # noqa

__all__ = ['Worker']

#: Name prefix of process in ps/top listings.
PSIDENT = '[Faust:Worker]'

TP_TYPES = (TP, TopicPartition)

W_SHUTDOWN_DUE_TO_MEMORY = '''\
Worker process is consuming {amount} of available memory.

Application will gracefully shutdown now to avoid an abrupt termination due 
an out of memory situation. If this happens frequently, consider
increasing the application's memory limit.
'''

logger = get_logger(__name__)


@formatter2
def format_log_arguments(
        arg: Any, record: logging.LogRecord) -> Any:  # pragma: no cover
    # This adds custom formatting to certain log messages.

    if arg and isinstance(arg, Mapping):
        first_k, first_v = next(iter(arg.items()))
        # Mapping of name to TopicT is changed to terminal table.
        if (isinstance(first_k, str) and isinstance(first_v, set) and
                isinstance(next(iter(first_v), None), TopicT)):
            return '\n' + terminal.logtable(
                sorted(arg.items()),
                title='Subscription',
                headers=['Topic', 'Descriptions'],
            )
        # Mapping where values are TopicPartition tuples are changed
        # to a terminal table.
        elif isinstance(first_v, TP_TYPES):
            return '\n' + terminal.logtable(
                [(k.topic, k.partition, v)
                 for k, v in sorted(arg.items())],
                title='Topic Partition Map',
                headers=['topic', 'partition', 'offset'],
            )
    elif arg and isinstance(arg, (set, list)):
        if 'Subscribed to topic' in record.msg:
            return '\n' + terminal.logtable(
                [
                    [str(v)] for v in sorted(arg)
                ],
                title='Final Subscription',
                headers=['topic name'],
            )
        elif isinstance(next(iter(arg)), TP_TYPES):
            # Sets/Lists of TopicPartition are converted to terminal table.
            return _partition_set_logtable(arg)

    elif arg and isinstance(arg, frozenset):
        if 'subscribed topics to' in record.msg:
            # aiokafka emits a frozenset of topics,
            # and we convert this to a terminal table.
            return '\n' + terminal.logtable(
                [
                    [str(v)] for v in sorted(arg)
                ],
                title='Requested Subscription',
                headers=['topic name'],
            )
        elif isinstance(next(iter(arg)), TP_TYPES):
            # Sets/Lists of TopicPartition are converted to terminal table.
            return _partition_set_logtable(arg)


def _partition_set_logtable(arg: Iterable[TP]) -> str:
    topics: Dict[str, Set[int]] = defaultdict(set)
    for tp in arg:
        topics[tp.topic].add(tp.partition)

    return '\n' + terminal.logtable(
        [(k, _repr_partition_set(v))
            for k, v in sorted(topics.items())],
        title='Topic Partition Set',
        headers=['topic', 'partitions'],
    )


def _repr_partition_set(s: Set[int]) -> str:
    """Convert set of partition numbers to human readable form.

    This will consolidate ranges of partitions to make them easier
    to read.

    Example:
        >>> partitions = {1, 2, 3, 7, 8, 9, 10, 34, 35, 36, 37, 38, 50}
        >>> _repr_partition_set(partitions)
        '{1-3, 7-10, 34-38, 50}'
    """
    elements = ', '.join(_iter_consecutive_numbers(sorted(s)))
    return f'{{{elements}}}'


def _iter_consecutive_numbers(s: Iterable[int]) -> Iterable[str]:
    """Find consecutive number ranges from an iterable of integers.

    The number ranges are represented as strings (e.g. ``"3-14"``)

    Example:
        >>> numbers = {1, 2, 3, 7, 8, 9, 10, 34, 35, 36, 37, 38, 50}
        >>> list(_iter_consecutive_numbers(numbers))
        [1-3, 7-10, 34-38, 50]
    """
    for numbers in consecutive_numbers(s):
        if len(numbers) > 1:
            yield f'{numbers[0]}-{numbers[-1]}'
        else:
            yield f'{numbers[0]}'


class Worker(mode.Worker):
    """Worker.

    See Also:
        This is a subclass of :class:`mode.Worker`.

    Usage:
        You can start a worker using:

            1) the :program:`faust worker` program.

            2) instantiating Worker programmatically and calling
               `execute_from_commandline()`::

                    >>> worker = Worker(app)
                    >>> worker.execute_from_commandline()

            3) or if you already have an event loop, calling ``await start``,
               but in that case *you are responsible for gracefully shutting
               down the event loop*::

                    async def start_worker(worker: Worker) -> None:
                        await worker.start()

                    def manage_loop():
                        loop = asyncio.get_event_loop()
                        worker = Worker(app, loop=loop)
                        try:
                            loop.run_until_complete(start_worker(worker)
                        finally:
                            worker.stop_and_shutdown_loop()

    Arguments:
        app: The Faust app to start.
        *services: Services to start with worker.
            This includes application instances to start.

        sensors (Iterable[SensorT]): List of sensors to include.
        debug (bool): Enables debugging mode [disabled by default].
        quiet (bool): Do not output anything to console [disabled by default].
        loglevel (Union[str, int]): Level to use for logging, can be string
            (one of: CRIT|ERROR|WARN|INFO|DEBUG), or integer.
        logfile (Union[str, IO]): Name of file or a stream to log to.
        stdout (IO): Standard out stream.
        stderr (IO): Standard err stream.
        blocking_timeout (float): When :attr:`debug` is enabled this
            sets the timeout for detecting that the event loop is blocked.
        workdir (Union[str, Path]): Custom working directory for the process
            that the worker will change into when started.
            This working directory change is permanent for the process,
            or until something else changes the working directory again.
        loop (asyncio.AbstractEventLoop): Custom event loop object.
    """

    logger = logger

    #: The Faust app started by this worker.
    app: AppT

    #: Additional sensors to add to the Faust app.
    sensors: Set[SensorT]

    #: Current working directory.
    #: Note that if passed as an argument to Worker, the worker
    #: will change to this directory when started.
    workdir: Path

    #: Class that displays a terminal progress spinner (see :pypi:`progress`).
    spinner: Optional[terminal.Spinner]

    #: Set by signal to avoid printing an OK status.
    _shutdown_immediately: bool = False

    #: Set when worker has initilized and started
    _on_startup_finished: Event

    #: OS process running the worker
    _worker_process: psutil.Process

    shutdown_memory_utilization_percent: float

    shutdown_memory_utilization_bytes: float

    def __init__(self,
                 app: AppT,
                 *services: ServiceT,
                 sensors: Iterable[SensorT] = None,
                 debug: bool = DEBUG,
                 quiet: bool = False,
                 loglevel: Union[str, int] = None,
                 logfile: Union[str, IO] = None,
                 stdout: IO = sys.stdout,
                 stderr: IO = sys.stderr,
                 blocking_timeout: float = BLOCKING_TIMEOUT,
                 workdir: Union[Path, str] = None,
                 console_port: int = CONSOLE_PORT,
                 loop: asyncio.AbstractEventLoop = None,
                 redirect_stdouts: bool = None,
                 redirect_stdouts_level: Severity = None,
                 logging_config: Dict = None,
                 shutdown_memory_utilization_percent: float = None,
                 shutdown_memory_utilization_bytes: float = None,
                 **kwargs: Any) -> None:
        self.app = app
        self.sensors = set(sensors or [])
        self.workdir = Path(workdir or Path.cwd())
        conf = app.conf
        self.shutdown_memory_utilization_percent = conf.worker_shutdown_memory_utilization_percent
        self.shutdown_memory_utilization_bytes = conf.worker_shutdown_memory_utilization_bytes
        if redirect_stdouts is None:
            redirect_stdouts = conf.worker_redirect_stdouts
        if redirect_stdouts_level is None:
            redirect_stdouts_level = (
                conf.worker_redirect_stdouts_level or logging.INFO)
        if logging_config is None:
            logging_config = app.conf.logging_config

        super().__init__(
            *services,
            debug=debug,
            quiet=quiet,
            loglevel=loglevel,
            logfile=logfile,
            loghandlers=app.conf.loghandlers,
            stdout=stdout,
            stderr=stderr,
            blocking_timeout=blocking_timeout,
            console_port=console_port,
            redirect_stdouts=redirect_stdouts,
            redirect_stdouts_level=redirect_stdouts_level,
            logging_config=logging_config,
            loop=loop,
            **kwargs)
        self.spinner = terminal.Spinner(file=self.stdout)
        self._on_startup_finished = Event()
        self._worker_process = psutil.Process(os.getpid())

    async def on_start(self) -> None:
        """Signal called every time the worker starts."""
        # This flag is set when running inside a worker process.
        self.app.in_worker = True

        await super().on_start()

    def _on_sigint(self) -> None:
        self._flag_as_shutdown_by_signal()
        super()._on_sigint()

    def _on_sigterm(self) -> None:
        self._flag_as_shutdown_by_signal()
        super()._on_sigterm()

    def _flag_as_shutdown_by_signal(self) -> None:
        self._shutdown_immediately = True
        if self.spinner:
            self.spinner.stop()

    async def maybe_start_blockdetection(self) -> None:
        """Start blocking detector if enabled."""
        if self.debug or FORCE_BLOCKING_TIMEOUT:
            self.log.info('Starting blocking detector with timeout %r',
                          self.blocking_timeout)
            await self.blocking_detector.maybe_start()

    async def on_startup_finished(self) -> None:
        """Signal called when worker has started."""
        self._on_startup_finished.set()
        if self._shutdown_immediately:
            return self._on_shutdown_immediately()
        # block detection started here after changelog stuff,
        # and blocking RocksDB bulk updates.
        await self.maybe_start_blockdetection()

        # end progress spinner, or it will mangle terminal output.
        self._on_startup_end_spinner()

    def _on_startup_end_spinner(self) -> None:
        if self.spinner:
            self.spinner.finish()
            if self.spinner.file.isatty():
                self.say(' 😊')
            else:
                self.say(' OK ^')
        else:
            self.log.info('Ready')

    def _on_shutdown_immediately(self) -> None:
        self.say('')  # make sure spinner newlines.

    def on_init_dependencies(self) -> Iterable[ServiceT]:
        """Return service dependencies that must start with the worker."""
        # App service is now a child of worker.
        self.app.beacon.reattach(self.beacon)
        # Transfer sensors to app
        for sensor in self.sensors:
            self.app.sensors.add(sensor)
        # Callback called once the app is running and fully
        # functional, we use it to e.g. print the "ready" message.
        self.app.on_startup_finished = self.on_startup_finished
        return chain(self.services, [self.app])

    async def on_first_start(self) -> None:
        """Signal called the first time the worker starts.

        First time, means this callback is not called if the
        worker is restarted by an exception being raised.
        """
        self.change_workdir(self.workdir)
        self.autodiscover()
        await self.default_on_first_start()

    def change_workdir(self, path: Path) -> None:
        """Change the current working directory (CWD)."""
        if path and path.absolute() != path.cwd().absolute():
            os.chdir(path.absolute())

    def autodiscover(self) -> None:
        """Autodiscover modules and files to find @agent decorators, etc."""
        if self.app.conf.autodiscover:
            self.app.discover()

    def _setproctitle(self, info: str, *, ident: str = PSIDENT) -> None:
        setproctitle(f'{ident} -{info}- {self._proc_ident()}')

    def _proc_ident(self) -> str:
        conf = self.app.conf
        return f'{conf.id} {self._proc_web_ident()} {conf.datadir.absolute()}'

    def _proc_web_ident(self) -> str:
        conf = self.app.conf
        if conf.web_transport.scheme == 'unix':
            return f'{conf.web_transport}'
        return f'-p {conf.web_port}'

    async def on_execute(self) -> None:
        """Signal called when the worker is about to start."""
        # This is called as soon as we start
        self._setproctitle('init')
        if self.spinner and self.spinner.file.isatty():
            self._say('starting➢ ', end='', flush=True)

    def on_worker_shutdown(self) -> None:
        """Signal called before the worker is shutting down."""
        self._setproctitle('stopping')
        if self.spinner and self.spinner.file.isatty():
            self.spinner.reset()
            self._say('stopping➢ ', end='', flush=True)

    def on_setup_root_logger(self, logger: logging.Logger, level: int) -> None:
        """Signal called when the root logger is being configured."""
        # This is used to set up the terminal progress spinner
        # so that it spins for every log message emitted.
        self._disable_spinner_if_level_below_WARN(level)
        self._setup_spinner_handler(logger, level)

    def _disable_spinner_if_level_below_WARN(self, level: int) -> None:
        if level and level < logging.WARN:
            self.spinner = None

    def _setup_spinner_handler(
            self, logger: logging.Logger, level: int) -> None:
        if self.spinner:
            logger.handlers[0].setLevel(level)
            logger.addHandler(
                terminal.SpinnerHandler(self.spinner, level=logging.DEBUG))
            logger.setLevel(logging.DEBUG)


    @mode.Service.task
    async def _stop_worker_when_high_memory_utilization(self) -> None:
        """Monitor worker's memory utilization.
        If memory utilization exceeds set threshold then gracefully stop 
        worker before we run out of memory.
        """
        await self.wait(self._on_startup_finished)
        while not self.should_stop:
            used_percent = self._worker_process.memory_percent()
            used_bytes = self._worker_process.memory_info().rss        
            if self.shutdown_memory_utilization_percent > 0:
                if used_percent > self.shutdown_memory_utilization_percent:
                    self.log.info(
                        W_SHUTDOWN_DUE_TO_MEMORY.format(
                            amount='{:.2%}'.format(used_percent)
                        )
                    )
                    await self.stop()
            if self.shutdown_memory_utilization_bytes > 0:
                if used_bytes > self.shutdown_memory_utilization_bytes:
                    self.log.info(
                        W_SHUTDOWN_DUE_TO_MEMORY.format(
                            amount='{:.2f} bytes'.format(used_bytes)
                        )
                    )
                    await self.stop()
            await self.sleep(15)


"""Events received in streams."""
import typing

from types import TracebackType
from typing import Any, Awaitable, Optional, Type, Union, cast

from faust.types import (
    AppT,
    ChannelT,
    CodecArg,
    EventT,
    HeadersArg,
    K,
    Message,
    MessageSentCallback,
    RecordMetadata,
    SchemaT,
    V,
)

if typing.TYPE_CHECKING:  # pragma: no cover
    from .app.base import App as _App
else:
    class _App: ...  # noqa

USE_EXISTING_KEY = object()
USE_EXISTING_VALUE = object()
USE_EXISTING_HEADERS = object()


class Event(EventT):
    """An event received on a channel.

    Notes:
        - Events have a key and a value::

            event.key, event.value

        - They also have a reference to the original message
          (if available), such as a Kafka record:

            event.message.offset

        - Iterating over channels/topics yields Event:

            async for event in channel:
                ...

        - Iterating over a stream (that in turn iterate over channel) yields
          Event.value::

            async for value in channel.stream()  # value is event.value
                ...

        - If you only have a Stream object, you can also access underlying
          events by using ``Stream.events``.

            For example:

            .. sourcecode:: python

                async for event in channel.stream.events():
                    ...

          Also commonly used for finding the "current event" related to
          a value in the stream:

          .. sourcecode:: python

              stream = channel.stream()
              async for event in stream.events():
                  event = stream.current_event
                  message = event.message
                  topic = event.message.topic

          You can retrieve the current event in a stream to:

              - Get access to the serialized key+value.
              - Get access to message properties like, what topic+partition
                the value was received on, or its offset.

          If you want access to both key and value, you should use
          ``stream.items()`` instead.

            .. sourcecode:: python

                async for key, value in stream.items():
                    ...

            ``stream.current_event`` can also be accessed but you must take
            extreme care you are using the correct stream object. Methods
            such as ``.group_by(key)`` and ``.through(topic)`` returns cloned
            stream objects, so in the example:

            The best way to access the current_event in an agent is
            to use the :class:`~contextvars.ContextVar`:

            .. sourcecode:: python

                from faust import current_event

                @app.agent(topic)
                async def process(stream):
                    async for value in stream:
                        event = current_event()
    """

    def __init__(self,
                 app: AppT,
                 key: K,
                 value: V,
                 headers: Optional[HeadersArg],
                 message: Message) -> None:
        self.app: AppT = app
        self.key: K = key
        self.value: V = value
        self.message: Message = message
        if headers is not None:
            if not isinstance(headers, dict):
                self.headers = dict(headers)
            else:
                self.headers = headers
        else:
            self.headers = {}

        self.acked: bool = False

    async def send(self,
                   channel: Union[str, ChannelT],
                   key: K = USE_EXISTING_KEY,
                   value: V = USE_EXISTING_VALUE,
                   partition: int = None,
                   timestamp: float = None,
                   headers: Any = USE_EXISTING_HEADERS,
                   schema: SchemaT = None,
                   key_serializer: CodecArg = None,
                   value_serializer: CodecArg = None,
                   callback: MessageSentCallback = None,
                   force: bool = False) -> Awaitable[RecordMetadata]:
        """Send object to channel."""
        if key is USE_EXISTING_KEY:
            key = self.key
        if value is USE_EXISTING_VALUE:
            value = self.value
        if headers is USE_EXISTING_HEADERS:
            headers = self.headers
        return await self._send(
            channel,
            key,
            value,
            partition,
            timestamp,
            headers,
            schema,
            key_serializer,
            value_serializer,
            callback,
            force=force,
        )

    async def forward(self,
                      channel: Union[str, ChannelT],
                      key: K = USE_EXISTING_KEY,
                      value: V = USE_EXISTING_VALUE,
                      partition: int = None,
                      timestamp: float = None,
                      headers: Any = USE_EXISTING_HEADERS,
                      schema: SchemaT = None,
                      key_serializer: CodecArg = None,
                      value_serializer: CodecArg = None,
                      callback: MessageSentCallback = None,
                      force: bool = False) -> Awaitable[RecordMetadata]:
        """Forward original message (will not be reserialized)."""
        if key is USE_EXISTING_KEY:
            key = self.message.key
        if value is USE_EXISTING_VALUE:
            value = self.message.value
        if headers is USE_EXISTING_HEADERS:
            headers = self.message.headers
            if not headers:
                headers = None
        return await self._send(
            channel,
            key,
            value,
            partition,
            timestamp,
            headers,
            schema,
            key_serializer,
            value_serializer,
            callback,
            force=force,
        )

    async def _send(self,
                    channel: Union[str, ChannelT],
                    key: K = None,
                    value: V = None,
                    partition: int = None,
                    timestamp: float = None,
                    headers: HeadersArg = None,
                    schema: SchemaT = None,
                    key_serializer: CodecArg = None,
                    value_serializer: CodecArg = None,
                    callback: MessageSentCallback = None,
                    force: bool = False) -> Awaitable[RecordMetadata]:
        return await cast(_App, self.app)._attachments.maybe_put(
            channel,
            key,
            value,
            partition,
            timestamp,
            headers,
            schema,
            key_serializer,
            value_serializer,
            callback,
            force=force,
        )

    def _attach(self,
                channel: Union[ChannelT, str],
                key: K = None,
                value: V = None,
                partition: int = None,
                timestamp: float = None,
                headers: HeadersArg = None,
                schema: SchemaT = None,
                key_serializer: CodecArg = None,
                value_serializer: CodecArg = None,
                callback: MessageSentCallback = None,
                ) -> Awaitable[RecordMetadata]:
        return cast(_App, self.app)._attachments.put(
            self.message,
            channel,
            key,
            value,
            partition=partition,
            timestamp=timestamp,
            headers=headers,
            schema=schema,
            key_serializer=key_serializer,
            value_serializer=value_serializer,
            callback=callback,
        )

    def ack(self) -> bool:
        """Acknowledge event as being processed by stream.

        When the last stream processor acks the message, the
        offset in the source topic will be marked as safe-to-commit,
        and the worker will commit and advance the committed offset.
        """
        return self.message.ack(self.app.consumer)

    def __repr__(self) -> str:
        return f'<{type(self).__name__}: k={self.key!r} v={self.value!r}>'

    async def __aenter__(self) -> EventT:
        return self

    async def __aexit__(self,
                        _exc_type: Type[BaseException] = None,
                        _exc_val: BaseException = None,
                        _exc_tb: TracebackType = None) -> Optional[bool]:
        self.ack()
        return None



from collections import deque
import pytest
from faust.utils.functional import (
    consecutive_numbers,
    deque_pushpopmax,
)


@pytest.mark.parametrize('numbers,expected', [
    ([1, 2, 3, 4, 6, 7, 8], [1, 2, 3, 4]),
    ([1, 4, 6, 8, 10], [1]),
    ([1], [1]),
    ([103, 104, 105, 106, 100000000000], [103, 104, 105, 106]),
])
def test_consecutive_numbers(numbers, expected):
    assert next(consecutive_numbers(numbers), None) == expected


def test_deque_pushpop_max():
    d = deque([])
    deque_pushpopmax(d, 1, max=None)
    assert d == deque([1])
    deque_pushpopmax(d, 2, max=3)
    assert d == deque([1, 2])
    deque_pushpopmax(d, 3, max=3)
    assert d == deque([1, 2, 3])
    deque_pushpopmax(d, 4, max=3)
    assert d == deque([2, 3, 4])
    for i in range(5, 100):
        deque_pushpopmax(d, i, max=3)
        assert d == deque([i - 2, i - 1, i])



"""Monitor using Statsd."""
import re
import typing

from typing import Any, Dict, Optional, Pattern, cast

from mode.utils.objects import cached_property

from faust import web
from faust.exceptions import ImproperlyConfigured
from faust.types import (
    AppT,
    CollectionT,
    EventT,
    Message,
    PendingMessage,
    RecordMetadata,
    StreamT,
    TP,
)
from faust.types.assignor import PartitionAssignorT
from faust.types.transports import ConsumerT, ProducerT

from .monitor import Monitor, TPOffsetMapping

try:
    import statsd
except ImportError:  # pragma: no cover
    statsd = None

if typing.TYPE_CHECKING:  # pragma: no cover
    from statsd import StatsClient
else:
    class StatsClient: ...  # noqa

__all__ = ['StatsdMonitor']

# This regular expression is used to generate stream ids in Statsd.
# It converts for example
#    "Stream: <Topic: withdrawals>"
# -> "Stream_Topic_withdrawals"
#
# See StatsdMonitor._normalize()
RE_NORMALIZE = re.compile(r'[\<\>:\s]+')
RE_NORMALIZE_SUBSTITUTION = '_'


class StatsdMonitor(Monitor):
    """Statsd Faust Sensor.

    This sensor, records statistics to Statsd along with computing metrics
    for the stats server
    """

    host: str
    port: int
    prefix: str

    def __init__(self,
                 host: str = 'localhost',
                 port: int = 8125,
                 prefix: str = 'faust-app',
                 rate: float = 1.0,
                 **kwargs: Any) -> None:
        self.host = host
        self.port = port
        self.prefix = prefix
        self.rate = rate
        if statsd is None:
            raise ImproperlyConfigured(
                'StatsMonitor requires `pip install statsd`.')
        super().__init__(**kwargs)

    def _new_statsd_client(self) -> StatsClient:
        return statsd.StatsClient(
            host=self.host, port=self.port, prefix=self.prefix)

    def on_message_in(self, tp: TP, offset: int, message: Message) -> None:
        """Call before message is delegated to streams."""
        super().on_message_in(tp, offset, message)

        self.client.incr('messages_received', rate=self.rate)
        self.client.incr('messages_active', rate=self.rate)
        self.client.incr(f'topic.{tp.topic}.messages_received', rate=self.rate)
        self.client.gauge(f'read_offset.{tp.topic}.{tp.partition}', offset)

    def on_stream_event_in(self, tp: TP, offset: int, stream: StreamT,
                           event: EventT) -> Optional[Dict]:
        """Call when stream starts processing an event."""
        state = super().on_stream_event_in(tp, offset, stream, event)
        self.client.incr('events', rate=self.rate)
        self.client.incr(
            f'stream.{self._stream_label(stream)}.events',
            rate=self.rate,
        )
        self.client.incr('events_active', rate=self.rate)
        return state

    def _stream_label(self, stream: StreamT) -> str:
        return self._normalize(
            stream.shortlabel.lstrip('Stream:'),
        ).strip('_').lower()

    def on_stream_event_out(self, tp: TP, offset: int, stream: StreamT,
                            event: EventT, state: Dict = None) -> None:
        """Call when stream is done processing an event."""
        super().on_stream_event_out(tp, offset, stream, event, state)
        self.client.decr('events_active', rate=self.rate)
        self.client.timing(
            'events_runtime',
            self.secs_to_ms(self.events_runtime[-1]),
            rate=self.rate)

    def on_message_out(self,
                       tp: TP,
                       offset: int,
                       message: Message) -> None:
        """Call when message is fully acknowledged and can be committed."""
        super().on_message_out(tp, offset, message)
        self.client.decr('messages_active', rate=self.rate)

    def on_table_get(self, table: CollectionT, key: Any) -> None:
        """Call when value in table is retrieved."""
        super().on_table_get(table, key)
        self.client.incr(f'table.{table.name}.keys_retrieved', rate=self.rate)

    def on_table_set(self, table: CollectionT, key: Any, value: Any) -> None:
        """Call when new value for key in table is set."""
        super().on_table_set(table, key, value)
        self.client.incr(f'table.{table.name}.keys_updated', rate=self.rate)

    def on_table_del(self, table: CollectionT, key: Any) -> None:
        """Call when key in a table is deleted."""
        super().on_table_del(table, key)
        self.client.incr(f'table.{table.name}.keys_deleted', rate=self.rate)

    def on_commit_completed(self, consumer: ConsumerT, state: Any) -> None:
        """Call when consumer commit offset operation completed."""
        super().on_commit_completed(consumer, state)
        self.client.timing(
            'commit_latency',
            self.ms_since(cast(float, state)),
            rate=self.rate)

    def on_send_initiated(self, producer: ProducerT, topic: str,
                          message: PendingMessage,
                          keysize: int, valsize: int) -> Any:
        """Call when message added to producer buffer."""
        self.client.incr(f'topic.{topic}.messages_sent', rate=self.rate)
        return super().on_send_initiated(
            producer, topic, message, keysize, valsize)

    def on_send_completed(self,
                          producer: ProducerT,
                          state: Any,
                          metadata: RecordMetadata) -> None:
        """Call when producer finished sending message."""
        super().on_send_completed(producer, state, metadata)
        self.client.incr('messages_sent', rate=self.rate)
        self.client.timing(
            'send_latency',
            self.ms_since(cast(float, state)),
            rate=self.rate)

    def on_send_error(self,
                      producer: ProducerT,
                      exc: BaseException,
                      state: Any) -> None:
        """Call when producer was unable to publish message."""
        super().on_send_error(producer, exc, state)
        self.client.incr('messages_sent_error', rate=self.rate)
        self.client.timing(
            'send_latency_for_error',
            self.ms_since(cast(float, state)),
            rate=self.rate)

    def on_assignment_error(self,
                            assignor: PartitionAssignorT,
                            state: Dict,
                            exc: BaseException) -> None:
        """Partition assignor did not complete assignor due to error."""
        super().on_assignment_error(assignor, state, exc)
        self.client.incr('assignments_error', rate=self.rate)
        self.client.timing(
            'assignment_latency',
            self.ms_since(state['time_start']),
            rate=self.rate)

    def on_assignment_completed(self,
                                assignor: PartitionAssignorT,
                                state: Dict) -> None:
        """Partition assignor completed assignment."""
        super().on_assignment_completed(assignor, state)
        self.client.incr('assignments_complete', rate=self.rate)
        self.client.timing(
            'assignment_latency',
            self.ms_since(state['time_start']),
            rate=self.rate)

    def on_rebalance_start(self, app: AppT) -> Dict:
        """Cluster rebalance in progress."""
        state = super().on_rebalance_start(app)
        self.client.incr('rebalances', rate=self.rate)
        return state

    def on_rebalance_return(self, app: AppT, state: Dict) -> None:
        """Consumer replied assignment is done to broker."""
        super().on_rebalance_return(app, state)
        self.client.decr('rebalances', rate=self.rate)
        self.client.incr('rebalances_recovering', rate=self.rate)
        self.client.timing(
            'rebalance_return_latency',
            self.ms_since(state['time_return']),
            rate=self.rate)

    def on_rebalance_end(self, app: AppT, state: Dict) -> None:
        """Cluster rebalance fully completed (including recovery)."""
        super().on_rebalance_end(app, state)
        self.client.decr('rebalances_recovering', rate=self.rate)
        self.client.timing(
            'rebalance_end_latency',
            self.ms_since(state['time_end']),
            rate=self.rate)

    def count(self, metric_name: str, count: int = 1) -> None:
        """Count metric by name."""
        super().count(metric_name, count=count)
        self.client.incr(metric_name, count=count, rate=self.rate)

    def on_tp_commit(self, tp_offsets: TPOffsetMapping) -> None:
        """Call when offset in topic partition is committed."""
        super().on_tp_commit(tp_offsets)
        for tp, offset in tp_offsets.items():
            metric_name = f'committed_offset.{tp.topic}.{tp.partition}'
            self.client.gauge(metric_name, offset)

    def track_tp_end_offset(self, tp: TP, offset: int) -> None:
        """Track new topic partition end offset for monitoring lags."""
        super().track_tp_end_offset(tp, offset)
        metric_name = f'end_offset.{tp.topic}.{tp.partition}'
        self.client.gauge(metric_name, offset)

    def on_web_request_end(self,
                           app: AppT,
                           request: web.Request,
                           response: Optional[web.Response],
                           state: Dict,
                           *,
                           view: web.View = None) -> None:
        """Web server finished working on request."""
        super().on_web_request_end(app, request, response, state, view=view)
        status_code = int(state['status_code'])
        self.client.incr(f'http_status_code.{status_code}', rate=self.rate)
        self.client.timing(
            'http_response_latency',
            self.ms_since(state['time_end']),
            rate=self.rate)

    def _normalize(self, name: str,
                   *,
                   pattern: Pattern = RE_NORMALIZE,
                   substitution: str = RE_NORMALIZE_SUBSTITUTION) -> str:
        return pattern.sub(substitution, name)

    @cached_property
    def client(self) -> StatsClient:
        """Return statsd client."""
        return self._new_statsd_client()



from collections import Counter
import pytest
from faust.tables.recovery import RebalanceAgain, Recovery, ServiceStopped
from faust.types import TP
from mode.utils.mocks import AsyncMock, Mock

TP1 = TP('foo', 6)
TP2 = TP('bar', 3)
TP3 = TP('baz', 1)
TP4 = TP('xuz', 0)


@pytest.fixture()
def tables():
    return Mock(name='tables')


@pytest.fixture()
def recovery(*, tables, app):
    return Recovery(app, tables)


class test_Recovery:

    @pytest.fixture()
    def table(self):
        return Mock(name='table')

    def test_init(self, *, recovery, tables):
        assert recovery.tables is tables
        assert recovery.signal_recovery_start
        assert recovery.signal_recovery_end
        assert recovery.signal_recovery_reset

    @pytest.mark.asyncio
    async def test_on_stop(self, *, recovery):
        recovery.flush_buffers = Mock()
        await recovery.on_stop()
        recovery.flush_buffers.assert_called_once_with()

    def test_add_active(self, *, recovery, table):
        recovery.add_active(table, TP1)
        assert TP1 in recovery.active_tps
        assert TP1 in recovery.actives_for_table[table]
        assert recovery.tp_to_table[TP1] is table
        assert recovery.active_offsets[TP1] is table.persisted_offset()
        recovery.revoke(TP1)
        assert TP1 not in recovery.active_offsets

    def test_add_standby(self, *, recovery, table):
        recovery.add_standby(table, TP1)
        assert TP1 in recovery.standby_tps
        assert TP1 in recovery.standbys_for_table[table]
        assert recovery.tp_to_table[TP1] is table
        assert recovery.standby_offsets[TP1] is table.persisted_offset()
        recovery.revoke(TP1)
        assert TP1 not in recovery.standby_offsets

    def test_on_partitions_revoked(self, *, recovery):
        recovery.signal_recovery_reset.clear()
        recovery.flush_buffers = Mock()

        recovery.on_partitions_revoked({TP1})

        recovery.flush_buffers.assert_called_once_with()
        assert recovery.signal_recovery_reset.is_set()

    @pytest.mark.asyncio
    async def test_on_rebalance(self, *, recovery, app, tables):
        recovery.signal_recovery_reset.set()
        app.assignor = Mock()
        app.assignor.assigned_standbys.return_value = {TP1}
        app.assignor.assigned_actives.return_value = {TP2}
        tables._changelogs = {
            TP1.topic: Mock(name='table1'),
            TP2.topic: Mock(name='table2'),
        }
        await recovery.on_rebalance({TP1, TP2, TP3}, {TP4}, {TP3})
        assert recovery.signal_recovery_start.is_set()
        assert not recovery.signal_recovery_reset.is_set()

        assert TP1 in recovery.standby_tps
        assert TP2 in recovery.active_tps

    @pytest.mark.asyncio
    async def test_on_rebalance__empty(self, *, recovery, app):
        app.assignor = Mock()
        app.assignor.assigned_standbys.return_value = set()
        app.assignor.assigned_actives.return_value = set()
        await recovery.on_rebalance(set(), set(), set())
        assert recovery.signal_recovery_start.is_set()

    @pytest.mark.asyncio
    async def test__resume_streams(self, *, recovery, tables, app):
        app.tables = tables
        app.on_rebalance_complete = Mock(send=AsyncMock())
        app.on_rebalance_end = Mock()
        app.flow_control = Mock()
        app._fetcher = Mock(maybe_start=AsyncMock())
        consumer = app.consumer = Mock()
        recovery._wait = AsyncMock()

        await recovery._resume_streams()
        app.on_rebalance_complete.send.assert_called_once_with()
        consumer.resume_flow.assert_called_once_with()
        app.flow_control.resume.assert_called_once_with()
        recovery._wait.assert_called_once_with(consumer.perform_seek())
        consumer.resume_partitions.assert_called_once_with(
            consumer.assignment())

        assert recovery.completed.is_set()
        app._fetcher.maybe_start.assert_called_once_with()
        app.tables.on_actives_ready.assert_called_once_with()
        app.tables.on_standbys_ready.assert_called_once_with()
        app.on_rebalance_end.assert_called_once_with()

        consumer.assignment.return_value = set()
        await recovery._resume_streams()

    @pytest.mark.asyncio
    async def test__wait(self, *, recovery):
        assert await self.assert_wait(
            recovery, stopped=False, done=None) is None

    @pytest.mark.asyncio
    async def test__wait__stopped(self, *, recovery):
        with pytest.raises(ServiceStopped):
            await self.assert_wait(recovery, stopped=True, done=None)

    @pytest.mark.asyncio
    async def test__wait__recovery_restart(self, *, recovery):
        with pytest.raises(RebalanceAgain):
            await self.assert_wait(
                recovery,
                stopped=False,
                done=recovery.signal_recovery_start)

    @pytest.mark.asyncio
    async def test__wait__recovery_reset(self, *, recovery):
        with pytest.raises(RebalanceAgain):
            await self.assert_wait(
                recovery,
                stopped=False,
                done=recovery.signal_recovery_reset)

    async def assert_wait(self, recovery,
                          stopped=False,
                          done=None):
        coro = Mock()
        recovery.wait_first = AsyncMock()
        recovery.wait_first.coro.return_value.stopped = stopped
        recovery.wait_first.coro.return_value.done = {done} if done else set()

        ret = await recovery._wait(coro)
        recovery.wait_first.assert_called_once_with(
            coro,
            recovery.signal_recovery_reset,
            recovery.signal_recovery_start,
        )
        return ret

    @pytest.mark.asyncio
    async def test_on_recovery_completed(self, *, recovery, tables, app):
        assignment = {TP1, TP2, TP3, TP4}
        consumer = app.consumer = Mock(
            name='consumer',
            perform_seek=AsyncMock(),
            assignment=Mock(return_value=assignment),
        )
        app.assignor = Mock(assigned_standbys=Mock(return_value={TP2}))
        recovery._is_changelog_tp = Mock(return_value=True)
        app._fetcher = Mock(maybe_start=AsyncMock())
        app.on_rebalance_complete = Mock(send=AsyncMock())
        app.on_rebalance_end = Mock()
        table1 = Mock(name='table1', on_recovery_completed=AsyncMock())
        table2 = Mock(name='table2', on_recovery_completed=AsyncMock())
        tables.values.return_value = [table1, table2]
        recovery.actives_for_table[table1] = {TP1}
        recovery.actives_for_table[table2] = {TP2}
        recovery.standbys_for_table[table1] = {TP3}
        recovery.standbys_for_table[table2] = {TP4}

        await recovery.on_recovery_completed()
        app.on_rebalance_complete.send.assert_called_once_with()
        table1.on_recovery_completed.assert_called_once_with({TP1}, {TP3})
        table2.on_recovery_completed.assert_called_once_with({TP2}, {TP4})
        consumer.perform_seek.assert_called_once_with()
        assert recovery.completed.is_set()
        consumer.resume_partitions.assert_called_once()
        app._fetcher.maybe_start.assert_called_once_with()
        tables.on_actives_ready.assert_called_once_with()
        app.on_rebalance_end.assert_called_once_with()

        assignment.clear()
        app.assignor.assigned_standbys.return_value = set()
        await recovery.on_recovery_completed()

        tables.values.return_value = []
        await recovery.on_recovery_completed()

    @pytest.mark.asyncio
    async def test__build_highwaters(self, *, recovery):
        tps = {TP1, TP2, TP3}
        dest = Counter({TP1: 103, TP4: 10})
        consumer = Mock(
            name='consumer',
            highwaters=AsyncMock(
                return_value={
                    TP1: 1001,
                    TP2: 0,
                    TP3: 202,
                },
            ),
        )
        await recovery._build_highwaters(consumer, tps, dest, 'title')
        assert dest == Counter({
            TP1: 1000,
            TP2: -1,
            TP3: 201,
        })

    @pytest.mark.asyncio
    async def test__build_offsets(self, *, recovery):
        tps = {TP1, TP2, TP3}
        dest = Counter({
            TP1: 300,
            TP2: 101,
            TP3: 2003,
        })
        consumer = Mock(
            name='consumer',
            earliest_offsets=AsyncMock(
                return_value={
                    TP1: 0,
                    TP2: 201,
                    TP3: 3003,
                },
            ),
        )
        await recovery._build_offsets(consumer, tps, dest, 'title')
        assert dest == Counter({
            TP1: 300,
            TP2: 200,
            TP3: 3002,
        })

    @pytest.mark.asyncio
    async def test__seek_offsets(self, *, recovery):
        consumer = Mock(
            name='consumer',
            seek_wait=AsyncMock(),
        )
        offsets = {
            TP1: -1,
            TP2: 1001,
            TP3: 2002,
        }
        tps = {TP1, TP2, TP3}

        await recovery._seek_offsets(consumer, tps, offsets, 'seek')
        consumer.seek_wait.assert_called_once_with({
            TP1: 0,
            TP2: 1001,
            TP3: 2002,
        })

    def test_flush_buffers(self, *, recovery):
        recovery.buffers.update({
            Mock(name='table1'): Mock(name='buffer1'),
            Mock(name='table2'): Mock(name='buffer2'),
        })
        recovery.flush_buffers()

        assert len(recovery.buffers) == 2
        for table, buffer in recovery.buffers.items():
            table.apply_changelog_batch.assert_called_once_with(buffer)
            buffer.clear.assert_called_once_with()

    def test_need_recovery__yes(self, *, recovery):
        self._setup_active_offsets(recovery)
        assert recovery.need_recovery()

    def test_need_recovery__no(self, *, recovery):
        self._setup_active_offsets(recovery)
        recovery.active_offsets = recovery.active_highwaters
        assert not recovery.need_recovery()

    def test_active_remaining(self, *, recovery):
        self._setup_active_offsets(recovery)
        assert recovery.active_remaining_total() == 2002

    def test_standby_remaining(self, *, recovery):
        self._setup_standby_offsets(recovery)
        assert recovery.standby_remaining_total() == 2002

    def test_active_stats(self, *, recovery):
        self._setup_active_offsets(recovery)
        assert recovery.active_stats() == {
            TP2: (3003, 2002, 1001),
            TP3: (4004, 3003, 1001),
        }

    def _setup_active_offsets(self, recovery):
        recovery.active_offsets = Counter({
            TP1: 1001,
            TP2: 2002,
            TP3: 3003,
        })
        recovery.active_highwaters = Counter({
            TP1: 1001,
            TP2: 3003,
            TP3: 4004,
        })

    def test_standby_stats(self, *, recovery):
        self._setup_standby_offsets(recovery)
        assert recovery.standby_stats() == {
            TP2: (3003, 2002, 1001),
            TP3: (4004, 3003, 1001),
        }

    def _setup_standby_offsets(self, recovery):
        recovery.standby_offsets = Counter({
            TP1: 1001,
            TP2: 2002,
            TP3: 3003,
        })
        recovery.standby_highwaters = Counter({
            TP1: 1001,
            TP2: 3003,
            TP3: 4004,
        })

    def test__is_changelog_tp(self, *, recovery, tables):
        tables.changelog_topics = {TP1.topic}
        assert recovery._is_changelog_tp(TP1)


@pytest.mark.parametrize('highwaters,offsets,needs_recovery,total,remaining', [
    ({TP1: 0, TP2: -1}, {TP1: -1, TP2: -1}, True, 1, {TP1: 1, TP2: 0}),
    ({TP1: -1, TP2: -1}, {TP1: -1, TP2: -1}, False, 0, {TP1: 0, TP2: 0}),
    ({TP1: 100, TP2: -1}, {TP1: -1, TP2: -1}, True, 101, {TP1: 101, TP2: 0}),
])
def test_recovery_from_offset_0(
        highwaters, offsets, needs_recovery, total, remaining, *,
        recovery):
    recovery.active_highwaters.update(highwaters)
    recovery.active_offsets.update(offsets)

    if needs_recovery:
        assert recovery.need_recovery()
    else:
        assert not recovery.need_recovery()
    assert recovery.active_remaining_total() == total
    if remaining:
        assert recovery.active_remaining() == remaining



import json
from random import random
from kafka import KafkaProducer

TOPIC = 'test'
KEY = 'score'


def publish_message(producer_instance, topic_name, key, value):
    try:
        key_bytes = bytes(key, encoding='utf-8')
        value_bytes = bytes(value, encoding='utf-8')
        producer_instance.send(topic_name, key=key_bytes, value=value_bytes)
        producer_instance.flush()
        print('Message published successfully.')
    except Exception as ex:
        print('Exception in publishing message')
        print(ex)


def connect_kafka_producer():
    _producer = None
    try:
        # host.docker.internal is how a docker container connects to the local
        # machine.
        # Don't use in production, this only works with Docker for Mac in
        # development
        _producer = KafkaProducer(
            bootstrap_servers=['host.docker.internal:9092'],
            api_version=(0, 10))
    except Exception as ex:
        print('Exception while connecting Kafka')
        print(ex)
    return _producer


if __name__ == '__main__':
    kafka_producer = connect_kafka_producer()
    for index in range(0, 10000):
        message = {
            'index': index,
            'value': round(random(), 2),
        }
        publish_message(kafka_producer, TOPIC, KEY, json.dumps(message))
    if kafka_producer is not None:
        kafka_producer.close()



import sys
import faust
sys.path.insert(0, '.')
from t.bench.base import Benchmark  # noqa

app = faust.App('bench-forward')
topic = app.topic(
    'bench-forward-1st',
    value_serializer='raw',
    value_type=int,
)
consume_topic = app.topic(
    'bench-forward-2nd',
    value_serializer='raw',
    value_type=int,
)

# Producer sends to topic1 (1st)
# Agent forwarder consumes from topic1 and forwards to topic2
# Benchmark agent consumes from topic2
#
#
# So we are measuring the time it takes to go through an
# additional agent.


@app.agent(topic)
async def forwarder(stream):
    i = 0
    async for event in stream.events():
        if i and not i % 10_000:
            print(f'Forwarded: {i}')
        await event.forward(consume_topic)
        i += 1


Benchmark(app, topic, consume_topic=consume_topic).install(__name__)



import os
import pytest
from faust.types._env import _getenv


def test_getenv_not_set():
    os.environ.pop('THIS_IS_NOT_SET', None)
    with pytest.raises(KeyError):
        _getenv('THIS_IS_NOT_SET')
